from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.preisstaffel import Preisstaffel
from bo4e.com.tarifpreisposition import Tarifpreisposition
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preistyp import Preistyp
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip

example_tarifpreisposition = Tarifpreisposition(
    preistyp=Preistyp.ENTGELT_ABLESUNG,
    einheit=Waehrungseinheit.EUR,
    bezugseinheit=Mengeneinheit.KWH,
    preisstaffeln=[
        Preisstaffel(
            einheitspreis=Decimal(40.0),
            staffelgrenze_von=Decimal(12.5),
            staffelgrenze_bis=Decimal(25.0),
        ),
    ],
)


class TestTarifpreisposition:
    @pytest.mark.parametrize(
        "tarifpreisposition, expected_json_dict",
        [
            pytest.param(
                example_tarifpreisposition,
                {
                    "preistyp": "ENTGELT_ABLESUNG",
                    "einheit": Waehrungseinheit.EUR,
                    "bezugseinheit": Mengeneinheit.KWH,
                    "preisstaffeln": [
                        {
                            "einheitspreis": Decimal("40"),
                            "sigmoidparameter": None,
                            "staffelgrenzeBis": Decimal("25"),
                            "staffelgrenzeVon": Decimal("12.5"),
                        }
                    ],
                    "mengeneinheitstaffel": None,
                },
                id="only required attributes",
            ),
            pytest.param(
                Tarifpreisposition(
                    preistyp=Preistyp.ENTGELT_ABLESUNG,
                    einheit=Waehrungseinheit.EUR,
                    bezugseinheit=Mengeneinheit.KWH,
                    preisstaffeln=[
                        Preisstaffel(
                            einheitspreis=Decimal(40.0),
                            staffelgrenze_von=Decimal(12.5),
                            staffelgrenze_bis=Decimal(25.0),
                        ),
                    ],
                    mengeneinheitstaffel=Mengeneinheit.STUECK,
                ),
                {
                    "preistyp": "ENTGELT_ABLESUNG",
                    "einheit": Waehrungseinheit.EUR,
                    "bezugseinheit": Mengeneinheit.KWH,
                    "preisstaffeln": [
                        {
                            "einheitspreis": Decimal("40"),
                            "sigmoidparameter": None,
                            "staffelgrenzeBis": Decimal("25"),
                            "staffelgrenzeVon": Decimal("12.5"),
                        }
                    ],
                    "mengeneinheitstaffel": Mengeneinheit.STUECK,
                },
                id="optional and required attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, tarifpreisposition: Tarifpreisposition, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Tarifpreisposition.
        """
        assert_serialization_roundtrip(tarifpreisposition, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tarifpreisposition()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)

    def test_tarifpreisposition_betraege_required(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tarifpreisposition(
                preistyp=Preistyp.ENTGELT_ABLESUNG,
                einheit=Waehrungseinheit.EUR,
                bezugseinheit=Mengeneinheit.KWH,
                preisstaffeln=[],
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Kontaktart(StrEnum):
    """
    Art des Kontaktes zwischen Geschäftspartnern.
    """

    ANSCHREIBEN = "ANSCHREIBEN"  #: Anschreiben
    TELEFONAT = "TELEFONAT"  #: Telefonat
    FAX = "FAX"  #: Fax
    E_MAIL = "E_MAIL"  #: Email
    SMS = "SMS"  #: SMS



"""
Contains Ausschreibungslos class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.com.ausschreibungsdetail import Ausschreibungsdetail
from bo4e.com.com import COM
from bo4e.com.menge import Menge
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.preismodell import Preismodell
from bo4e.enum.rechnungslegung import Rechnungslegung
from bo4e.enum.sparte import Sparte
from bo4e.enum.vertragsform import Vertragsform


class Ausschreibungslos(COM):
    """
    Eine Komponente zur Abbildung einzelner Lose einer Ausschreibung

    .. raw:: html

        <object data="../_static/images/bo4e/com/Ausschreibungslos.svg" type="image/svg+xml"></object>

    .. HINT::
        `Ausschreibungslos JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Ausschreibungslos.json>`_

    """

    # required attributes
    #: Laufende Nummer des Loses
    losnummer: str
    #: Bezeichnung der Ausschreibung
    bezeichnung: str
    #: Bezeichnung der Preismodelle in Ausschreibungen für die Energielieferung
    preismodell: Preismodell

    #: Unterscheidungsmöglichkeiten für die Sparte
    energieart: Sparte
    #: Aufzählung der Möglichkeiten zur Rechnungslegung in Ausschreibungen
    wunsch_rechnungslegung: Rechnungslegung
    #: Aufzählung der Möglichkeiten zu Vertragsformen in Ausschreibungen
    wunsch_vertragsform: Vertragsform
    #: Name des Lizenzpartners
    betreut_durch: str
    #: Anzahl der Lieferstellen in dieser Ausschreibung
    anzahl_lieferstellen: int

    #: Die ausgeschriebenen Lieferstellen
    lieferstellen: conlist(Ausschreibungsdetail, min_items=1)  # type: ignore[valid-type]

    #: Zeitraum, für den die in diesem Los enthaltenen Lieferstellen beliefert werden sollen
    lieferzeitraum: Zeitraum

    # optional attributes
    #: Bemerkung des Kunden zum Los
    bemerkung: Optional[str] = None
    #: Gibt den Gesamtjahresverbrauch (z.B. in kWh) aller in diesem Los enthaltenen Lieferstellen an
    gesamt_menge: Optional[Menge] = None
    #: Mindesmenge Toleranzband (kWh, %)
    wunsch_mindestmenge: Optional[Menge] = None
    #: Maximalmenge Toleranzband (kWh, %)
    wunsch_maximalmenge: Optional[Menge] = None

    wiederholungsintervall: Optional[Zeitraum] = None
    """
    In welchem Intervall die Angebotsabgabe wiederholt werden darf.
    Angabe nur gesetzt für die 2. Phase bei öffentlich-rechtlichen Ausschreibungen
    """

    #: Kundenwunsch zur Kündigungsfrist in der Ausschreibung
    wunsch_kuendingungsfrist: Optional[Zeitraum] = None
    #: Kundenwunsch zum Zahlungsziel in der Ausschreibung
    wunsch_zahlungsziel: Optional[Zeitraum] = None



"""
Contains Angebotsvariante and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.com.angebotsteil import Angebotsteil
from bo4e.com.betrag import Betrag
from bo4e.com.com import COM
from bo4e.com.menge import Menge
from bo4e.enum.angebotsstatus import Angebotsstatus


class Angebotsvariante(COM):
    """
    Führt die verschiedenen Ausprägungen der Angebotsberechnung auf

    .. raw:: html

        <object data="../_static/images/bo4e/com/Angebotsvariante.svg" type="image/svg+xml"></object>

    .. HINT::
        `Angebotsvariante JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Angebotsvariante.json>`_

    """

    # required attributes
    #: Gibt den Status eines Angebotes an.
    angebotsstatus: Angebotsstatus

    #: Datum der Erstellung der Angebotsvariante
    erstellungsdatum: datetime

    #: Bis zu diesem Zeitpunkt gilt die Angebotsvariante
    bindefrist: datetime

    teile: conlist(Angebotsteil, min_items=1)  # type: ignore[valid-type]
    """
    Angebotsteile werden im einfachsten Fall für eine Marktlokation oder Lieferstellenadresse erzeugt.
    Hier werden die Mengen und Gesamtkosten aller Angebotspositionen zusammengefasst.
    Eine Variante besteht mindestens aus einem Angebotsteil.
    """

    # optional attributes
    #: Aufsummierte Wirkarbeitsmenge aller Angebotsteile
    gesamtmenge: Optional[Menge] = None
    # todo: write a validator for this: https://github.com/Hochfrequenz/BO4E-python/issues/320
    #: Aufsummierte Kosten aller Angebotsteile
    gesamtkosten: Optional[Betrag] = None



"""
Contains Fremdkosten class and corresponding marshmallow schema for de-/serialization
"""
from typing import List, Optional

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.betrag import Betrag
from bo4e.com.fremdkostenblock import Fremdkostenblock
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.botyp import BoTyp

# pylint: disable=too-few-public-methods


class Fremdkosten(Geschaeftsobjekt):
    """
    Mit diesem BO werden die Fremdkosten, beispielsweise für eine Angebotserstellung oder eine Rechnungsprüfung,
    übertragen.
    Die Fremdkosten enthalten dabei alle Kostenblöcke, die von anderen Marktteilnehmern oder Instanzen erhoben werden.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Fremdkosten.svg" type="image/svg+xml"></object>

    .. HINT::
        `Fremdkosten JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Fremdkosten.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.FREMDKOSTEN
    #: Für diesen Zeitraum wurden die Kosten ermittelt
    gueltigkeit: Zeitraum
    # optional attributes
    #: Die Gesamtsumme über alle Kostenblöcke und -positionen
    summe_kosten: Optional[Betrag] = None
    #: In Kostenblöcken werden Kostenpositionen zusammengefasst. Beispiele: Netzkosten, Umlagen, Steuern etc
    kostenbloecke: Optional[List[Fremdkostenblock]] = None



# pylint: disable=missing-module-docstring, line-too-long
from bo4e.enum.strenum import StrEnum


class Rechnungstyp(StrEnum):
    """
    Abbildung verschiedener Rechnungstypen zur Kennzeichnung von Rechnungen
    """

    ENDKUNDENRECHNUNG = (
        "ENDKUNDENRECHNUNG"  #: Eine Rechnung vom Lieferanten an einen Endkunden über die Lieferung von Energie.
    )
    NETZNUTZUNGSRECHNUNG = "NETZNUTZUNGSRECHNUNG"  #: Eine Rechnung vom Netzbetreiber an den Netznutzer. (i.d.R. der Lieferant) über die Netznutzung.
    MEHRMINDERMENGENRECHNUNG = "MEHRMINDERMENGENRECHNUNG"  #: Eine Rechnung vom Netzbetreiber an den Netznutzer. (i.d.R. der Lieferant) zur Abrechnung von Mengen-Differenzen zwischen Bilanzierung und Messung.
    MESSSTELLENBETRIEBSRECHNUNG = (
        "MESSSTELLENBETRIEBSRECHNUNG"  #: Rechnung eines Messstellenbetreibers an den Messkunden.
    )
    BESCHAFFUNGSRECHNUNG = "BESCHAFFUNGSRECHNUNG"  #: Rechnungen zwischen einem  Händler und Einkäufer von Energie.
    AUSGLEICHSENERGIERECHNUNG = "AUSGLEICHSENERGIERECHNUNG"  #: Rechnung an den Verursacher von Ausgleichsenergie.



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Regionskriteriumtyp(StrEnum):
    """
    Klassifizierung der Kriterien für eine regionale Eingrenzung.
    """

    BUNDESLANDKENNZIFFER = "BUNDESLANDKENNZIFFER"  #: offizielle Bundeslandkennziffer
    BUNDESLAND_NAME = "BUNDESLAND_NAME"  #: Bundesland Name
    MARKTGEBIET_NUMMER = "MARKTGEBIET_NUMMER"  #: offizielle Marktgebiet-Codenummer
    MARKTGEBIET_NAME = "MARKTGEBIET_NAME"  #: Marktgebiet Name
    REGELGEBIET_NUMMER = "REGELGEBIET_NUMMER"  #: offizielle Regelgebiet Nummer
    REGELGEBIET_NAME = "REGELGEBIET_NAME"  #: Regelgebiet Name
    NETZ_STROM = "NETZ_STROM"  #: Identifikation des Netzes Strom
    NETZ_GAS = "NETZ_GAS"  #: Identifikation des Netzes Gas
    NETZBETREIBER_NUMMER_STROM = "NETZBETREIBER_NUMMER_STROM"  #: offizielle Netzbetreiber-Codenummer Strom
    NETZBETREIBER_NUMMER_GAS = "NETZBETREIBER_NUMMER_GAS"  #: offizielle Netzbetreiber-Codenummer Gas
    NETZBETREIBER_NAME_STROM = "NETZBETREIBER_NAME_STROM"  #: Netzbetreiber Name Strom
    NETZBETREIBER_NAME_GAS = "NETZBETREIBER_NAME_GAS"  #: Netzbetreiber Name Gas
    BILANZIERUNGS_GEBIET_NUMMER = (
        "BILANZIERUNGS_GEBIET_NUMMER"  #: Strom: Bilanzierungsgebietsnummer, Gas: Netzkontonummer
    )
    MSB_NUMMER = "MSB_NUMMER"  #: offizielle Messstellenbetreiber-Codenummer
    MSB_NAME = "MSB_NAME"  #: Name des MSB
    VERSORGER_NUMMER = "VERSORGER_NUMMER"  #: offizielle Lieferanten-Codenummer eines Versorgers
    VERSORGER_NAME = "VERSORGER_NAME"  #: Name eines Versorgers
    GRUNDVERSORGER_NUMMER_STROM = (
        "GRUNDVERSORGER_NUMMER_STROM"  #: offizielle Lieferanten-Codenummer des Strom-Grundversorgers
    )
    GRUNDVERSORGER_NAME_STROM = "GRUNDVERSORGER_NAME_STROM"  #: Name des Strom-Grundversorgers
    GRUNDVERSORGER_NUMMER_GAS = (
        "GRUNDVERSORGER_NUMMER_GAS"  #: offizielle Lieferanten-Codenummer des Gas-Grundversorgers
    )
    GRUNDVERSORGER_NAME_GAS = "GRUNDVERSORGER_NAME_GAS"  #: Name des Gas-Grundversorgers
    KREIS_NAME = "KREIS_NAME"  #: Kreis
    KREISKENNZIFFER = "KREISKENNZIFFER"  #: offizielle Kreiskennziffer
    GEMEINDE_NAME = "GEMEINDE_NAME"  #: Gemeinde
    GEMEINDEKENNZIFFER = "GEMEINDEKENNZIFFER"  #: offizielle Gemeindekennziffer
    POSTLEITZAHL = "POSTLEITZAHL"  #: Postleitzahl
    ORT = "ORT"  #: Ort
    POSTORT = "POSTORT"  #: Kriterium bestehend aus Postleitzahl und Ort
    EINWOHNERZAHL_GEMEINDE = "EINWOHNERZAHL_GEMEINDE"  #: Einwohnerzahl Gemeinde
    EINWOHNERZAHL_ORT = "EINWOHNERZAHL_ORT"  #: Einwohnerzahl Ort
    KM_UMKREIS = "KM_UMKREIS"  #: km Umkreis
    BUNDESWEIT = "BUNDESWEIT"  #: Bundesweite Betrachtung
    PLZ_BEREICH = "PLZ_BEREICH"  #: Postleitzahlenbereich



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Tariftyp(StrEnum):
    """
    Zur Differenzierung von Grund/Ersatzversorgungstarifen und sonstigen angebotenen Tarifen.
    """

    GRUND_ERSATZVERSORGUNG = "GRUND_ERSATZVERSORGUNG"  #: Grund- und Ersatzversorgung
    GRUNDVERSORGUNG = "GRUNDVERSORGUNG"  #: Grundversorgung
    ERSATZVERSORGUNG = "ERSATZVERSORGUNG"  #: Ersatzversorgung
    SONDERTARIF = "SONDERTARIF"  #: Sondertarif



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Lokationstyp(StrEnum):
    """
    Gibt an, ob es sich um eine Markt- oder Messlokation handelt.
    """

    MALO = "MALO"  #: Marktlokation
    MELO = "MELO"  #: Messlokation



"""
Contains RegionalePreisgarantie class
and corresponding marshmallow schema for de-/serialization
"""


from bo4e.com.preisgarantie import Preisgarantie
from bo4e.com.regionalegueltigkeit import RegionaleGueltigkeit

# pylint: disable=too-few-public-methods


class RegionalePreisgarantie(Preisgarantie):
    """
    Abbildung einer Preisgarantie mit regionaler Abgrenzung

    .. raw:: html

        <object data="../_static/images/bo4e/com/RegionalePreisgarantie.svg" type="image/svg+xml"></object>

    .. HINT::
        `RegionalePreisgarantie JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/RegionalePreisgarantie.json>`_

    """

    # required attributes
    #: Regionale Eingrenzung der Preisgarantie.
    regionale_gueltigkeit: RegionaleGueltigkeit



import datetime

import pytest
from pydantic import ValidationError

from bo4e.bo.tarifpreisblatt import Tarifpreisblatt
from bo4e.com.tarifberechnungsparameter import Tarifberechnungsparameter
from bo4e.com.tarifeinschraenkung import Tarifeinschraenkung
from bo4e.enum.kundentyp import Kundentyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.tarifmerkmal import Tarifmerkmal
from bo4e.enum.tariftyp import Tariftyp
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_aufabschlag import example_aufabschlag
from tests.test_energiemix import example_energiemix
from tests.test_geraeteeigenschaften import example_geraeteeigenschaften
from tests.test_marktteilnehmer import example_marktteilnehmer
from tests.test_preisgarantie import example_preisgarantie
from tests.test_preisposition import example_preisposition
from tests.test_tarifpreisposition import example_tarifpreisposition
from tests.test_vertragskonditionen import example_vertragskonditionen
from tests.test_zeitraum import example_zeitraum


class TestTarifpreisblatt:
    @pytest.mark.parametrize(
        "tarifpreisblatt",
        [
            pytest.param(
                Tarifpreisblatt(
                    bezeichnung="foo",
                    anbietername="der beste stromanbieter",
                    sparte=Sparte.STROM,
                    kundentypen=[Kundentyp.PRIVAT, Kundentyp.GEWERBE],
                    tarifart=Tarifart.MEHRTARIF,
                    tariftyp=Tariftyp.GRUND_ERSATZVERSORGUNG,
                    tarifmerkmale=[Tarifmerkmal.HEIZSTROM],
                    website="https://foo.inv",
                    bemerkung="super billig aber auch super dreckig",
                    vertragskonditionen=example_vertragskonditionen,
                    zeitliche_gueltigkeit=example_zeitraum,
                    energiemix=example_energiemix,
                    anbieter=example_marktteilnehmer,
                    # ^^ above is all copy pasted from Tarifinfo BO
                    # vv below are the attributes of tarifpreisblatt
                    preisstand=datetime.datetime(2022, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
                    berechnungsparameter=Tarifberechnungsparameter(),
                    tarif_auf_abschlaege=[example_aufabschlag],
                    tarifpreise=[example_tarifpreisposition],
                    preisgarantie=example_preisgarantie,
                    tarifeinschraenkung=Tarifeinschraenkung(),
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifpreisblatt: Tarifpreisblatt) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarifpreisblatt)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tarifpreisblatt()  # type: ignore[call-arg]
        assert "11 validation error" in str(excinfo.value)  # 8 from tarifinfo + 3 from tarifpreisblatt



"""
Contains RegionaleGueltigkeit class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.com.com import COM
from bo4e.com.kriteriumwert import KriteriumWert
from bo4e.enum.gueltigkeitstyp import Gueltigkeitstyp


class RegionaleGueltigkeit(COM):
    """
    Mit dieser Komponente können regionale Gültigkeiten, z.B. für Tarife, Zu- und Abschläge und Preise definiert werden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/RegionaleGueltigkeit.svg" type="image/svg+xml"></object>

    .. HINT::
        `RegionaleGueltigkeit JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/RegionaleGueltigkeit.json>`_

    """

    # required attributes
    gueltigkeitstyp: Gueltigkeitstyp  #: Unterscheidung ob Positivliste oder Negativliste übertragen wird
    kriteriums_werte: conlist(  # type: ignore[valid-type]
        KriteriumWert, min_items=1
    )  #: Hier stehen die Kriterien, die die regionale Gültigkeit festlegen



"""
Contains Lastgang and LastgangKompakt class
and corresponding marshmallow schema for de-/serialization
"""
from typing import List, Optional

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist, constr

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.tagesvektor import Tagesvektor
from bo4e.com.zeitintervall import Zeitintervall
from bo4e.com.zeitreihenwert import Zeitreihenwert
from bo4e.enum.botyp import BoTyp
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.sparte import Sparte
from bo4e.validators import OBIS_PATTERN


class _LastgangBody(Geschaeftsobjekt):
    """
    The LastgangBody is a mixin that contains the "body" of a Lastgang that is used in both the :class:`Lastgang` as
    well as :class:`LastgangKompakt`.
    """

    #: Angabe, ob es sich um einen Gas- oder Stromlastgang handelt
    sparte: Sparte

    #: Eindeutige Nummer der Marktlokation bzw der Messlokation, zu der der Lastgang gehört
    lokations_id: str

    #: Marktlokation oder Messlokation
    lokationstyp: str
    # todo: implement a lokations-id + lokationstyp cross check (such that lokationstyp malo checks for valid malo id)
    # https://github.com/Hochfrequenz/BO4E-python/issues/321

    #: Definition der gemessenen Größe anhand ihrer Einheit
    messgroesse: Mengeneinheit

    # optional attributes
    #: Versionsnummer des Lastgangs
    version: Optional[str] = None
    #: Die OBIS-Kennzahl für den Wert, die festlegt, welche Größe mit dem Stand gemeldet wird, z.B. '1-0:1.8.1'
    obis_kennzahl: Optional[constr(strict=True, regex=OBIS_PATTERN)] = None  # type: ignore[valid-type]


# pylint: disable=too-many-instance-attributes, too-few-public-methods


class LastgangKompakt(_LastgangBody):
    """
    Modell zur Abbildung eines kompakten Lastganges.
    In diesem Modell werden die Messwerte in Form von Tagesvektoren mit fester Anzahl von Werten übertragen.
    Daher ist dieses BO nur zur Übertragung von äquidistanten Messwertverläufen geeignet.
    """

    # required attributes
    bo_typ: BoTyp = BoTyp.LASTGANG_KOMPAKT

    #: Angabe des Rasters innerhalb aller Tagesvektoren dieses Lastgangs
    zeitintervall: Zeitintervall
    # todo: implement a cross check that this zeitintervall is actually the one used in tagesvektoren
    # https://github.com/Hochfrequenz/BO4E-python/issues/322

    #: Die im Lastgang enthaltenen Messwerte in Form von Tagesvektoren
    tagesvektoren: List[Tagesvektor]


# pylint: disable=too-many-instance-attributes, too-few-public-methods


class Lastgang(_LastgangBody):
    """
    Modell zur Abbildung eines Lastganges;
    In diesem Modell werden die Messwerte mit einem vollständigen Zeitintervall angegeben und es bietet daher eine hohe
    Flexibilität in der Übertragung jeglicher zeitlich veränderlicher Messgrössen.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Lastgang.svg" type="image/svg+xml"></object>

    .. HINT::
        `Lastgang JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Lastgang.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.LASTGANG

    #: Die im Lastgang enthaltenen Messwerte
    werte: conlist(Zeitreihenwert, min_items=1)  # type: ignore[valid-type]



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Messpreistyp(StrEnum):
    """
    Festlegung, welcher Typ von Messung mit einem Preis belegt wird
    """

    MESSPREIS_G2_5 = "MESSPREIS_G2_5"  #: MESSPREIS_G2_5
    MESSPREIS_G4 = "MESSPREIS_G4"  #: MESSPREIS_G4
    MESSPREIS_G6 = "MESSPREIS_G6"  #: MESSPREIS_G6
    MESSPREIS_G10 = "MESSPREIS_G10"  #: MESSPREIS_G10
    MESSPREIS_G16 = "MESSPREIS_G16"  #: MESSPREIS_G16
    MESSPREIS_G25 = "MESSPREIS_G25"  #: MESSPREIS_G25
    MESSPREIS_G40 = "MESSPREIS_G40"  #: MESSPREIS_G40
    ELEKTRONISCHER_AUFSATZ = "ELEKTRONISCHER_AUFSATZ"  #: ELEKTRONISCHER_AUFSATZ
    SMART_METER_MESSPREIS_G2_5 = "SMART_METER_MESSPREIS_G2_5"  #: SMART_METER_MESSPREIS_G2_5
    SMART_METER_MESSPREIS_G4 = "SMART_METER_MESSPREIS_G4"  #: SMART_METER_MESSPREIS_G4
    SMART_METER_MESSPREIS_G6 = "SMART_METER_MESSPREIS_G6"  #: SMART_METER_MESSPREIS_G6
    SMART_METER_MESSPREIS_G10 = "SMART_METER_MESSPREIS_G10"  #: SMART_METER_MESSPREIS_G10
    SMART_METER_MESSPREIS_G16 = "SMART_METER_MESSPREIS_G16"  #: SMART_METER_MESSPREIS_G16
    SMART_METER_MESSPREIS_G25 = "SMART_METER_MESSPREIS_G25"  #: SMART_METER_MESSPREIS_G25
    SMART_METER_MESSPREIS_G40 = "SMART_METER_MESSPREIS_G40"  #: SMART_METER_MESSPREIS_G40
    VERRECHNUNGSPREIS_ET_WECHSEL = "VERRECHNUNGSPREIS_ET_WECHSEL"  #: VERRECHNUNGSPREIS_ET_WECHSEL
    VERRECHNUNGSPREIS_ET_DREH = "VERRECHNUNGSPREIS_ET_DREH"  #: VERRECHNUNGSPREIS_ET_DREH
    VERRECHNUNGSPREIS_ZT_WECHSEL = "VERRECHNUNGSPREIS_ZT_WECHSEL"  #: VERRECHNUNGSPREIS_ZT_WECHSEL
    VERRECHNUNGSPREIS_ZT_DREH = "VERRECHNUNGSPREIS_ZT_DREH"  #: VERRECHNUNGSPREIS_ZT_DREH
    VERRECHNUNGSPREIS_L_ET = "VERRECHNUNGSPREIS_L_ET"  #: VERRECHNUNGSPREIS_L_ET
    VERRECHNUNGSPREIS_L_ZT = "VERRECHNUNGSPREIS_L_ZT"  #: VERRECHNUNGSPREIS_L_ZT
    VERRECHNUNGSPREIS_SM = "VERRECHNUNGSPREIS_SM"  #: VERRECHNUNGSPREIS_SM
    AUFSCHLAG_WANDLER = "AUFSCHLAG_WANDLER"  #: AUFSCHLAG_WANDLER
    AUFSCHLAG_TARIFSCHALTUNG = "AUFSCHLAG_TARIFSCHALTUNG"  #: AUFSCHLAG_TARIFSCHALTUNG



"""
Contains Geraet class
and corresponding marshmallow schema for de-/serialization
"""
from typing import Optional

from bo4e.com.com import COM
from bo4e.com.geraeteeigenschaften import Geraeteeigenschaften

# pylint: disable=too-few-public-methods


class Geraet(COM):
    """
    Mit dieser Komponente werden alle Geräte modelliert, die keine Zähler sind.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Geraet.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geraet JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Geraet.json>`_

    """

    # optional attributes
    #: Die auf dem Gerät aufgedruckte Nummer, die vom MSB vergeben wird.
    geraetenummer: Optional[str] = None
    #: Festlegung der Eigenschaften des Gerätes. Z.B. Wandler MS/NS.
    geraeteeigenschaften: Optional[Geraeteeigenschaften] = None



"""
Contains Marktgebietinfo class
and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods


class MarktgebietInfo(COM):
    """
    Informationen zum Marktgebiet im Gas.

    .. raw:: html

        <object data="../_static/images/bo4e/com/MarktgebietInfo.svg" type="image/svg+xml"></object>

    .. HINT::
        `MarktgebietInfo JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/MarktgebietInfo.json>`_

    """

    # required attributes
    marktgebiet: str  #: Der Name des Marktgebietes
    marktgebietcode: str  #: Die standardisierte Codenummer des Marktgebietes



"""
Contains Angebotsposition class
and corresponding marshmallow schema for de-/serialization
"""
from typing import Optional

from bo4e.com.betrag import Betrag
from bo4e.com.com import COM
from bo4e.com.menge import Menge
from bo4e.com.preis import Preis

# pylint: disable=too-few-public-methods


class Angebotsposition(COM):
    """
    Unterhalb von Angebotsteilen sind die Angebotspositionen eingebunden.
    Hier werden die angebotenen Bestandteile einzeln aufgeführt. Beispiel:
    Positionsmenge: 4000 kWh
    Positionspreis: 24,56 ct/kWh
    Positionskosten: 982,40 EUR

    .. raw:: html

        <object data="../_static/images/bo4e/com/Angebotsposition.svg" type="image/svg+xml"></object>

    .. HINT::
        `Angebotsposition JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Angebotsposition.json>`_

    """

    # required attributes
    #: Bezeichnung der jeweiligen Position des Angebotsteils
    positionsbezeichnung: str
    #: Preis pro Einheit/Stückpreis des angebotenen Artikels.
    positionspreis: Preis

    # optional attributes
    #: Menge des angebotenen Artikels (z.B. Wirkarbeit in kWh), in dieser Angebotsposition
    positionsmenge: Optional[Menge] = None
    #: Kosten (positionspreis * positionsmenge) für diese Angebotsposition
    positionskosten: Optional[Betrag] = None

    # for a preis = menge*times validation we first need to resolve
    # https://github.com/Hochfrequenz/BO4E-python/issues/126



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Ausschreibungsportal(StrEnum):
    """
    Aufzählung der unterstützten Ausschreibungsportale.
    """

    ENPORTAL = "ENPORTAL"  #: enPORTAL
    ENERGIE_AGENTUR = "ENERGIE_AGENTUR"  #: EnergieAgentur.NRW
    BMWI = "BMWI"  #: BMWI-Ausschreibungen
    ENERGIE_HANDELSPLATZ = "ENERGIE_HANDELSPLATZ"  #: energie-handelsplatz.de
    BUND = "BUND"  #: BUND.DE
    VERA_ONLINE = "VERA_ONLINE"  #: vera_online.de
    ISPEX = "ISPEX"  #: ispex.de
    ENERGIEMARKTPLATZ = "ENERGIEMARKTPLATZ"  #: energiemarktplatz.de
    EVERGABE = "EVERGABE"  #: evergabe.de
    DTAD = "DTAD"  #: dtad.de



"""
Contains Geschaeftspartner class
and corresponding marshmallow schema for de-/serialization
"""
# pylint: disable=too-many-instance-attributes, too-few-public-methods
from typing import List, Optional

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.adresse import Adresse
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.kontaktart import Kontaktart


class Geschaeftspartner(Geschaeftsobjekt):
    """
    Mit diesem Objekt können Geschäftspartner übertragen werden.
    Sowohl Unternehmen, als auch Privatpersonen können Geschäftspartner sein.
    Hinweis: Marktteilnehmer haben ein eigenes BO, welches sich von diesem BO ableitet.
    Hier sollte daher keine Zuordnung zu Marktrollen erfolgen.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Geschaeftspartner.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geschaeftspartner JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Geschaeftspartner.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.GESCHAEFTSPARTNER
    name1: str
    """
    Erster Teil des Namens.
    Hier kann der Firmenname oder bei Privatpersonen beispielsweise der Nachname dagestellt werden.
    Beispiele: Yellow Strom GmbH oder Hagen
    """
    # todo: replace name1/2/3 with something more readable. no one wants to deal with that. maybe serialize as name1/2/3
    # but resolve to readable python fields under the hood

    gewerbekennzeichnung: bool
    """
    Kennzeichnung ob es sich um einen Gewerbe/Unternehmen (gewerbeKennzeichnung = true)
    oder eine Privatperson handelt. (gewerbeKennzeichnung = false)
    """
    #: Rollen, die die Geschäftspartner inne haben (z.B. Interessent, Kunde)
    geschaeftspartnerrolle: List[Geschaeftspartnerrolle]
    # todo: rename to plural

    # optional attributes
    #: Die Anrede für den GePa, Z.B. "Herr"
    anrede: Optional[Anrede] = None
    name2: Optional[str] = None
    """
    Zweiter Teil des Namens.
    Hier kann der eine Erweiterung zum Firmennamen oder bei Privatpersonen beispielsweise der Vorname dagestellt werden.
    Beispiele: Bereich Süd oder Nina
    """

    name3: Optional[str] = None
    """
    Dritter Teil des Namens.
    Hier können weitere Ergänzungen zum Firmennamen oder bei Privatpersonen Zusätze zum Namen dagestellt werden.
    Beispiele: und Afrika oder Sängerin
    """
    #: Handelsregisternummer des Geschäftspartners
    hrnummer: Optional[str] = None
    #: Amtsgericht bzw Handelsregistergericht, das die Handelsregisternummer herausgegeben hat
    amtsgericht: Optional[str] = None
    #: Bevorzugte Kontaktwege des Geschäftspartners
    kontaktweg: Optional[List[Kontaktart]] = []
    #: Die Steuer-ID des Geschäftspartners; Beispiel: "DE 813281825"
    umsatzsteuer_id: Optional[str] = None
    #: Die Gläubiger-ID welche im Zahlungsverkehr verwendet wird; Z.B. "DE 47116789"
    glaeubiger_id: Optional[str] = None
    #: E-Mail-Adresse des Ansprechpartners. Z.B. "info@hochfrequenz.de"
    e_mail_adresse: Optional[str] = None
    #: Internetseite des Marktpartners
    website: Optional[str] = None
    #: Adressen der Geschäftspartner, an denen sich der Hauptsitz befindet
    partneradresse: Optional[Adresse] = None  # todo: is it plural or not? the docs are bad



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.kriteriumwert import KriteriumWert
from bo4e.com.regionalegueltigkeit import RegionaleGueltigkeit
from bo4e.com.regionalepreisstaffel import RegionalePreisstaffel
from bo4e.com.regionaletarifpreisposition import RegionaleTarifpreisposition
from bo4e.enum.gueltigkeitstyp import Gueltigkeitstyp
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preistyp import Preistyp
from bo4e.enum.tarifregionskriterium import Tarifregionskriterium
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_sigmoidparameter import example_sigmoidparameter

example_regionale_tarifpreisposition = RegionaleTarifpreisposition(
    preistyp=Preistyp.ARBEITSPREIS_NT,
    einheit=Waehrungseinheit.EUR,
    bezugseinheit=Mengeneinheit.KWH,
    mengeneinheitstaffel=Mengeneinheit.WH,
    preisstaffeln=[
        RegionalePreisstaffel(
            einheitspreis=Decimal(40.0),
            staffelgrenze_von=Decimal(12.5),
            staffelgrenze_bis=Decimal(25.0),
            sigmoidparameter=example_sigmoidparameter,
            regionale_gueltigkeit=RegionaleGueltigkeit(
                gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
                kriteriums_werte=[KriteriumWert(kriterium=Tarifregionskriterium.POSTLEITZAHL, wert="01069")],
            ),
        ),
    ],
)


class TestRegionaleTarifpreisPosition:
    @pytest.mark.parametrize(
        "regionale_tarifpreis_position, expected_json_dict",
        [
            pytest.param(
                example_regionale_tarifpreisposition,
                {
                    "bezugseinheit": Mengeneinheit.KWH,
                    "preisstaffeln": [
                        {
                            "regionaleGueltigkeit": {
                                "gueltigkeitstyp": "NUR_IN",
                                "kriteriumsWerte": [{"kriterium": Tarifregionskriterium.POSTLEITZAHL, "wert": "01069"}],
                            },
                            "einheitspreis": Decimal("40"),
                            "sigmoidparameter": {
                                "A": Decimal("1"),
                                "B": Decimal("2"),
                                "C": Decimal("3"),
                                "D": Decimal("4"),
                            },
                            "staffelgrenzeVon": Decimal("12.5"),
                            "staffelgrenzeBis": Decimal("25"),
                        }
                    ],
                    "einheit": Waehrungseinheit.EUR,
                    "mengeneinheitstaffel": Mengeneinheit.WH,
                    "preistyp": "ARBEITSPREIS_NT",
                },
                id="all attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, regionale_tarifpreis_position: RegionaleTarifpreisposition, expected_json_dict: Dict[str, Any]
    ) -> None:
        assert_serialization_roundtrip(regionale_tarifpreis_position, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = RegionaleTarifpreisposition()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.preisstaffel import Preisstaffel
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_sigmoidparameter import example_sigmoidparameter

example_preisstaffel = Preisstaffel(
    einheitspreis=Decimal(40.0), staffelgrenze_von=Decimal(12.5), staffelgrenze_bis=Decimal(25.0)
)


class TestPreisstaffel:
    @pytest.mark.parametrize(
        "preisstaffel, expected_json_dict",
        [
            pytest.param(
                Preisstaffel(
                    einheitspreis=Decimal(40.0),
                    staffelgrenze_von=Decimal(12.5),
                    staffelgrenze_bis=Decimal(25.0),
                    sigmoidparameter=example_sigmoidparameter,
                ),
                {
                    "einheitspreis": Decimal("40"),
                    "sigmoidparameter": {"A": Decimal("1"), "B": Decimal("2"), "C": Decimal("3"), "D": Decimal("4")},
                    "staffelgrenzeVon": Decimal("12.5"),
                    "staffelgrenzeBis": Decimal("25"),
                },
                id="all attributes",
            ),
            pytest.param(
                example_preisstaffel,
                {
                    "einheitspreis": Decimal("40"),
                    "staffelgrenzeVon": Decimal("12.5"),
                    "staffelgrenzeBis": Decimal("25"),
                    "sigmoidparameter": None,
                },
                id="only required params",
            ),
        ],
    )
    def test_serialization_roundtrip(self, preisstaffel: Preisstaffel, expected_json_dict: Dict[str, Any]) -> None:
        """
        Test de-/serialisation of Preisstaffel.
        """
        assert_serialization_roundtrip(preisstaffel, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Preisstaffel()  # type: ignore[call-arg]

        assert "3 validation errors" in str(excinfo.value)

    @pytest.mark.parametrize(
        "not_a_sigmoid_parameter",
        [
            pytest.param(17),  # not a sigmoid parameter instance
            pytest.param("foo"),  # not a sigmoid parameter instance
        ],
    )
    def test_failing_validation(self, not_a_sigmoid_parameter: Any) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Preisstaffel(
                einheitspreis=Decimal(40.0),
                staffelgrenze_von=Decimal(12.5),
                staffelgrenze_bis=Decimal(25.0),
                sigmoidparameter=not_a_sigmoid_parameter,
            )

        assert "1 validation error" in str(excinfo.value)
        assert "sigmoidparameter" in str(excinfo.value)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class ArtikelId(StrEnum):
    """
    Liste von Artikel-IDs, z.B. für standardisierte vom BDEW herausgegebene Artikel,
    die im Strommarkt die BDEW-Artikelnummer ablösen
    """

    # von allen enums ist das hier wohl das mit den unpassendsten Namen ;)

    ARTIKEL_2017001 = "2-01-7-001"  #: Unterbrechung der Anschlussnutzung in der regulären Arbeitszeit (€/Auftrag)
    ARTIKEL_2017002 = "2-01-7-002"  #: Wiederherstellung der Anschlussnutzung in der regulären Arbeitszeit (€/Auftrag)
    ARTIKEL_2017003 = "2-01-7-003"  #: Erfolglose Unterbrechung (€/Auftrag)
    #: Stornierung eines Auftrages zur Unterbrechung der Anschlussnutzung bis zum Vortag der Sperrung (€/Auftrag)
    ARTIKEL_2017004 = "2-01-7-004"
    #: Stornierung eines Auftrages zur Unterbrechung der Anschlussnutzung am Tag der Sperrung (€/Auftrag)
    ARTIKEL_2017005 = "2-01-7-005"
    #: Wiederherstellung der Anschlussnutzung außerhalb der regulären Arbeitszeit (€/Auftrag)
    ARTIKEL_2017006 = "2-01-7-006"
    ARTIKEL_2020001 = "2-02-0-001"  #: Verzugskosten pauschal (€/Fall)
    ARTIKEL_2020002 = "2-02-0-002"  #: Verzugskosten variabel (€)



import pytest
from pydantic import ValidationError

from bo4e.com.regionaleraufabschlag import RegionalerAufAbschlag
from bo4e.com.tarifeinschraenkung import Tarifeinschraenkung
from bo4e.enum.aufabschlagstyp import AufAbschlagstyp
from bo4e.enum.aufabschlagsziel import AufAbschlagsziel
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_energiemix import example_energiemix
from tests.test_preisgarantie import example_preisgarantie
from tests.test_regionalepreisstaffel import example_regionale_preisstaffel
from tests.test_vertragskonditionen import example_vertragskonditionen
from tests.test_zeitraum import example_zeitraum

example_regionaler_auf_abschlag = RegionalerAufAbschlag(
    bezeichnung="Foo",
    beschreibung="Bar",
    auf_abschlagstyp=AufAbschlagstyp.RELATIV,
    auf_abschlagsziel=AufAbschlagsziel.ARBEITSPREIS_HT,
    einheit=Waehrungseinheit.CT,
    website="https://www.hochfrequenz.de",
    zusatzprodukte=["ein standmixer", "ein thermomix"],
    voraussetzungen=["lecker essen", "mit ökostrom gekocht"],
    tarifnamensaenderungen="Super-Duper Tarif",
    staffeln=[example_regionale_preisstaffel],
    gueltigkeitszeitraum=example_zeitraum,
    energiemixaenderung=example_energiemix,
    vertagskonditionsaenderung=example_vertragskonditionen,
    garantieaenderung=example_preisgarantie,
    einschraenkungsaenderung=Tarifeinschraenkung(),
)


class TestRegionalerAufAbschlag:
    @pytest.mark.parametrize(
        "regionaler_auf_abschlag",
        [
            pytest.param(
                example_regionaler_auf_abschlag,
                id="maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, regionaler_auf_abschlag: RegionalerAufAbschlag) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(regionaler_auf_abschlag)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = RegionalerAufAbschlag()  # type: ignore[call-arg]
        assert "2 validation errors" in str(excinfo.value)



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.menge import Menge
from bo4e.enum.mengeneinheit import Mengeneinheit

example_menge = Menge(wert=Decimal(3.41), einheit=Mengeneinheit.MWH)
# see issue https://github.com/Hochfrequenz/BO4E-python/issues/249
example_menge_dict = {
    "wert": Decimal("3.410000000000000142108547152020037174224853515625"),
    "einheit": Mengeneinheit.MWH,
}


class TestMenge:
    def test_menge(self) -> None:
        """
        Test de-/serialisation of Menge (only has required attributes).
        """

        json_string = example_menge.json(by_alias=True, ensure_ascii=False)

        assert "3.41" in json_string
        assert "MWH" in json_string

        menge_deserialized = Menge.parse_raw(json_string)

        assert isinstance(menge_deserialized.wert, Decimal)
        assert menge_deserialized.wert == Decimal(3.41)
        assert isinstance(menge_deserialized.einheit, Mengeneinheit)
        assert menge_deserialized.einheit == Mengeneinheit.MWH

    def test_wrong_datatype(self) -> None:
        """
        A string "3.14" would be casted to decimal from pydantic therefore no validation error would occure in this case.
        """
        with pytest.raises(ValidationError) as excinfo:
            _ = Menge(wert="hallo", einheit=Mengeneinheit.MWH)  # type: ignore[arg-type]

        assert "wert" in str(excinfo.value)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Menge(wert=Decimal(3.14))  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Tarifart(StrEnum):
    """
    Die Tarifart wird verwendet zur Charakterisierung von Zählern und daraus resultierenden Tarifen.
    """

    EINTARIF = "EINTARIF"  #: Eintarif
    ZWEITARIF = "ZWEITARIF"  #: Zweitarif
    MEHRTARIF = "MEHRTARIF"  #: Mehrtarif
    SMART_METER = "SMART_METER"  #: Smart Meter Tarif
    LEISTUNGSGEMESSEN = "LEISTUNGSGEMESSEN"  #: Leistungsgemessener Tarif



import pytest
from pydantic import ValidationError

from bo4e.bo.region import Region
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_regionskriterium import example_regionskriterium


class TestRegion:
    @pytest.mark.parametrize(
        "region",
        [
            pytest.param(
                Region(
                    bezeichnung="Bikini Bottom",
                    positiv_liste=[example_regionskriterium],
                    negativ_liste=[example_regionskriterium],
                ),
                id="max attributes",
            ),
            pytest.param(
                Region(bezeichnung="Bikini Bottom", positiv_liste=[example_regionskriterium]), id="min attributes"
            ),
        ],
    )
    def test_serialization_roundtrip(self, region: Region) -> None:
        assert_serialization_roundtrip(region)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Region()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)

    def test_region_positiv_liste_required_and_negativ_liste_not_required(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Region(
                bezeichnung="Bikini Bottom",
                positiv_liste=[],
                negativ_liste=[],
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.angebotsvariante import Angebotsvariante
from bo4e.enum.angebotsstatus import Angebotsstatus
from bo4e.enum.mengeneinheit import Mengeneinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_angebotsteil import example_angebotsteil, example_angebotsteil_json
from tests.test_betrag import example_betrag, example_betrag_json
from tests.test_menge import example_menge

# can be imported by other tests
example_angebotsvariante = Angebotsvariante(
    angebotsstatus=Angebotsstatus.NACHGEFASST,
    bindefrist=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
    erstellungsdatum=datetime(2021, 12, 22, 0, 0, 0, tzinfo=timezone.utc),
    teile=[example_angebotsteil],
)


class TestAngebotsvariante:
    @pytest.mark.parametrize(
        "angebotsvariante, expected_json_dict",
        [
            pytest.param(
                example_angebotsvariante,
                {
                    "gesamtmenge": None,
                    "angebotsstatus": Angebotsstatus.NACHGEFASST,
                    "erstellungsdatum": datetime(2021, 12, 22, 0, 0, tzinfo=timezone.utc),
                    "gesamtkosten": None,
                    "bindefrist": datetime(2022, 2, 1, 0, 0, tzinfo=timezone.utc),
                    "teile": [example_angebotsteil_json],
                },
                id="minimal attributes",
            ),
            pytest.param(
                Angebotsvariante(
                    angebotsstatus=Angebotsstatus.NACHGEFASST,
                    bindefrist=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
                    erstellungsdatum=datetime(2021, 12, 22, 0, 0, 0, tzinfo=timezone.utc),
                    teile=[example_angebotsteil],
                    gesamtmenge=example_menge,
                    gesamtkosten=example_betrag,
                ),
                {
                    "gesamtmenge": {
                        "einheit": Mengeneinheit.MWH,
                        "wert": Decimal("3.410000000000000142108547152020037174224853515625"),
                    },
                    # this is a problem for https://github.com/Hochfrequenz/BO4E-python/issues/249
                    # I just reused the example_menge but don't attempt to fix it in the context of the Angebotsvariante
                    "angebotsstatus": Angebotsstatus.NACHGEFASST,
                    "erstellungsdatum": datetime(2021, 12, 22, 0, 0, tzinfo=timezone.utc),
                    "gesamtkosten": example_betrag_json,
                    "bindefrist": datetime(2022, 2, 1, 0, 0, tzinfo=timezone.utc),
                    "teile": [example_angebotsteil_json],
                },
                id="max attributes",  # = min + menge and betrag
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, angebotsvariante: Angebotsvariante, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation roundtrip.
        """
        assert_serialization_roundtrip(angebotsvariante, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Angebotsvariante()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.betrag import Betrag
from bo4e.enum.waehrungscode import Waehrungscode
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip

example_betrag = Betrag(
    waehrung=Waehrungscode.EUR,
    wert=Decimal(12.5),
)

example_betrag_json = {"wert": Decimal("12.5"), "waehrung": Waehrungseinheit.EUR}


class TestBetrag:
    @pytest.mark.parametrize(
        "betrag, expected_json_dict",
        [
            pytest.param(example_betrag, example_betrag_json),
        ],
    )
    def test_regionskriterium_serialization_roundtrip(self, betrag: Betrag, expected_json_dict: Dict[str, Any]) -> None:
        """
        Test de-/serialisation of Regionskriterium with minimal attributes.
        """
        assert_serialization_roundtrip(betrag, expected_json_dict)

    def test_regionskriterium_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Betrag()  # type: ignore[call-arg]
        assert "2 validation errors" in str(excinfo.value)



"""
Contains Kosten class and corresponding marshmallow schema for de-/serialization
"""
from typing import List, Optional

# pylint: disable=too-many-instance-attributes, too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.betrag import Betrag
from bo4e.com.kostenblock import Kostenblock
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.botyp import BoTyp
from bo4e.enum.kostenklasse import Kostenklasse


class Kosten(Geschaeftsobjekt):
    """
    Dieses BO wird zur Übertagung von hierarchischen Kostenstrukturen verwendet.
    Die Kosten werden dabei in Kostenblöcke und diese wiederum in Kostenpositionen strukturiert.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Kosten.svg" type="image/svg+xml"></object>

    .. HINT::
        `Kosten JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Kosten.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.KOSTEN
    #: Klasse der Kosten, beispielsweise Fremdkosten
    kostenklasse: Kostenklasse
    #: Für diesen Zeitraum wurden die Kosten ermittelt
    gueltigkeit: Zeitraum
    #: In Kostenblöcken werden Kostenpositionen zusammengefasst. Beispiele: Netzkosten, Umlagen, Steuern etc
    kostenbloecke: conlist(Kostenblock, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Die Gesamtsumme über alle Kostenblöcke und -positionen
    summe_kosten: Optional[List[Betrag]] = None



from datetime import datetime, timezone
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.preisgarantie import Preisgarantie
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.preisgarantietyp import Preisgarantietyp
from tests.serialization_helper import assert_serialization_roundtrip

example_preisgarantie = Preisgarantie(
    preisgarantietyp=Preisgarantietyp.NUR_ENERGIEPREIS,
    zeitliche_gueltigkeit=Zeitraum(
        startdatum=datetime(2020, 1, 1, tzinfo=timezone.utc),
        enddatum=datetime(2020, 4, 1, tzinfo=timezone.utc),
    ),
)


class TestPreisgarantie:
    @pytest.mark.parametrize(
        "preisgarantie, expected_json_dict",
        [
            pytest.param(
                example_preisgarantie,
                {
                    "beschreibung": None,
                    "preisgarantietyp": "NUR_ENERGIEPREIS",
                    "zeitlicheGueltigkeit": {
                        "startdatum": datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
                        "endzeitpunkt": None,
                        "einheit": None,
                        "enddatum": datetime(2020, 4, 1, 0, 0, tzinfo=timezone.utc),
                        "startzeitpunkt": None,
                        "dauer": None,
                    },
                },
            ),
        ],
    )
    def test_preisgarantie_required_attributes(
        self, preisgarantie: Preisgarantie, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Preisgarantie with minimal attributes.
        """
        assert_serialization_roundtrip(preisgarantie, expected_json_dict)

    def test_preisgarantie_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Preisgarantie()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Marktrolle(StrEnum):
    """
    Diese Rollen kann ein Marktteilnehmer einnehmen.
    """

    NB = "NB"  #: Verteilnetzbetreiber
    LF = "LF"  #: Lieferant
    MSB = "MSB"  #: Messstellenbetreiber
    DL = "DL"  #: Dienstleister
    BKV = "BKV"  #: Bilanzkreisverantwortlicher,
    BKO = "BKO"  #: Bilanzkoordinator / Marktgebietsverantwortlicher
    UENB = "UENB"  #: Übertragungsnetzbetreiber
    KUNDE_SELBST_NN = "KUNDE_SELBST_NN"  #: Kunden, die Netznutzungsentgelte selbst zahlen
    MGV = "MGV"  #: Marktgebietsverantwortlicher
    EIV = "EIV"  #: Einsatzverantwortlicher
    RB = "RB"  #: Registerbetreiber
    KUNDE = "KUNDE"  #: Endkunde
    INTERESSENT = "INTERESSENT"  #: Interessent
    BTR = "BTR"  #: Betreiber einer technischen Ressource



"""
Contains Zeitintervall class
and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM
from bo4e.enum.zeiteinheit import Zeiteinheit

# pylint: disable=too-few-public-methods


class Zeitintervall(COM):
    """
    Abbildung für ein Zeitintervall. Die Abbildung eines Zeitintervalls.
    Z.B. zur Anwendung als Raster in äquidistanten Zeitreihen/Lastgängen, beispielsweise 15 Minuten.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zeitintervall.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zeitintervall JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Zeitintervall.json>`_

    """

    # required attributes
    wert: int
    """
    Die Anzahl der Zeiteinheiten innerhalb  des Intervalls
    """
    zeiteinheit: Zeiteinheit
    """
    Die Einheit des Zeitintervalls
    """



"""
Contains RegionaleTarifpreisposition class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.com.com import COM
from bo4e.com.regionalepreisstaffel import RegionalePreisstaffel
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preistyp import Preistyp
from bo4e.enum.waehrungseinheit import Waehrungseinheit


class RegionaleTarifpreisposition(COM):
    """
    Mit dieser Komponente können Tarifpreise verschiedener Typen im Zusammenhang mit regionalen Gültigkeiten abgebildet
    werden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/RegionaleTarifpreisposition.svg" type="image/svg+xml"></object>

    .. HINT::
        `RegionaleTarifpreisposition JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/RegionaleTarifpreisposition.json>`_

    """

    # required attributes
    #: Angabe des Preistypes (z.B. Grundpreis)
    preistyp: Preistyp
    #: Einheit des Preises (z.B. EURO)
    einheit: Waehrungseinheit
    #: Größe, auf die sich die Einheit bezieht, beispielsweise kWh, Jahr
    bezugseinheit: Mengeneinheit
    #: Hier sind die Staffeln mit ihren Preisangaben und regionalen Gültigkeiten definiert
    preisstaffeln: conlist(RegionalePreisstaffel, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Gibt an, nach welcher Menge die vorgenannte Einschränkung erfolgt (z.B. Jahresstromverbrauch in kWh)
    mengeneinheitstaffel: Optional[Mengeneinheit] = None



from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.verbrauch import Verbrauch
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.wertermittlungsverfahren import Wertermittlungsverfahren
from tests.serialization_helper import assert_serialization_roundtrip

example_verbrauch = Verbrauch(
    wert=Decimal(40),
    obis_kennzahl="1-0:1.8.1",
    mengeneinheit=Mengeneinheit.KWH,
    wertermittlungsverfahren=Wertermittlungsverfahren.MESSUNG,
)


class TestVerbrauch:
    @pytest.mark.parametrize(
        "verbrauch, expected_json_dict",
        [
            pytest.param(
                Verbrauch(
                    wert=Decimal(40),
                    startdatum=datetime(2021, 12, 1, 0, 0, 0).replace(tzinfo=timezone.utc),
                    enddatum=datetime(2021, 12, 2, 0, 0, 0).replace(tzinfo=timezone.utc),
                    obis_kennzahl="1-0:1.8.1",
                    mengeneinheit=Mengeneinheit.KWH,
                    wertermittlungsverfahren=Wertermittlungsverfahren.MESSUNG,
                ),
                {
                    "startdatum": datetime(2021, 12, 1, 0, 0, tzinfo=timezone.utc),
                    "wert": Decimal("40"),
                    "mengeneinheit": Mengeneinheit.KWH,
                    "enddatum": datetime(2021, 12, 2, 0, 0, tzinfo=timezone.utc),
                    "wertermittlungsverfahren": Wertermittlungsverfahren.MESSUNG,
                    "obisKennzahl": "1-0:1.8.1",
                },
            ),
            pytest.param(
                example_verbrauch,
                {
                    "wert": Decimal("40"),
                    "mengeneinheit": Mengeneinheit.KWH,
                    "wertermittlungsverfahren": Wertermittlungsverfahren.MESSUNG,
                    "startdatum": None,
                    "enddatum": None,
                    "obisKennzahl": "1-0:1.8.1",
                },
            ),
        ],
    )
    def test_serialization_roundtrip(self, verbrauch: Verbrauch, expected_json_dict: Dict[str, Any]) -> None:
        """
        Test de-/serialisation of Verbrauch.
        """
        assert_serialization_roundtrip(verbrauch, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Verbrauch()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)

    @pytest.mark.parametrize(
        "not_a_valid_obis",
        [
            pytest.param("foo"),  # not a obis instance
        ],
    )
    def test_failing_validation_obis(self, not_a_valid_obis: str) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Verbrauch(
                obis_kennzahl=not_a_valid_obis,
                wert=Decimal(40),
                startdatum=datetime(2021, 12, 1, 0, 0, 0).replace(tzinfo=timezone.utc),
                enddatum=datetime(2021, 12, 2, 0, 0, 0).replace(tzinfo=timezone.utc),
                mengeneinheit=Mengeneinheit.KWH,
                wertermittlungsverfahren=Wertermittlungsverfahren.MESSUNG,
            )

        assert "1 validation error" in str(excinfo.value)
        assert "obisKennzahl" in str(excinfo.value)
        assert "string does not match regex" in str(excinfo.value)

    def test_failing_validation_end_later_than_start(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Verbrauch(
                obis_kennzahl="1-0:1.8.1",
                wert=Decimal(40),
                startdatum=datetime(2021, 12, 2, 0, 0, 0).replace(tzinfo=timezone.utc),
                enddatum=datetime(2021, 12, 1, 0, 0, 0).replace(tzinfo=timezone.utc),
                mengeneinheit=Mengeneinheit.KWH,
                wertermittlungsverfahren=Wertermittlungsverfahren.MESSUNG,
            )
        assert "has to be later than the start" in str(excinfo.value)



"""
Contains Energieherkunft class
and corresponding marshmallow schema for de-/serialization
"""
# pylint: disable=no-name-in-module
from pydantic import condecimal

from bo4e.com.com import COM
from bo4e.enum.erzeugungsart import Erzeugungsart

# pylint: disable=too-few-public-methods


class Energieherkunft(COM):
    """
    Abbildung einer Energieherkunft

    .. raw:: html

        <object data="../_static/images/bo4e/com/Energieherkunft.svg" type="image/svg+xml"></object>

    .. HINT::
        `Energieherkunft JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Energieherkunft.json>`_

    """

    # required attributes
    #: Art der Erzeugung der Energie.
    erzeugungsart: Erzeugungsart
    #: Prozentualer Anteil der jeweiligen Erzeugungsart.
    anteil_prozent: condecimal(gt=0, lt=100)  # type: ignore[valid-type]



"""
Contains PreisblattNetnutzung class and corresponding marshmallow schema for de-/serialization
"""


from bo4e.bo.preisblatt import Preisblatt
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.botyp import BoTyp
from bo4e.enum.kundengruppe import Kundengruppe
from bo4e.enum.netzebene import Netzebene

# pylint: disable=too-few-public-methods


class PreisblattNetznutzung(Preisblatt):
    """
    Die Variante des Preisblattmodells zur Abbildung der Netznutzungspreise

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattNetznutzung.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattNetznutzung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/PreisblattNetznutzung.json>`_

    """

    bo_typ: BoTyp = BoTyp.PREISBLATTNETZNUTZUNG
    # required attributes (additional to those of Preisblatt)
    #: Die Preise gelten für Marktlokationen der angebebenen Bilanzierungsmethode
    bilanzierungsmethode: Bilanzierungsmethode
    #: Die Preise gelten für Marktlokationen in der angebebenen Netzebene
    netzebene: Netzebene
    kundengruppe: Kundengruppe

    # there are no optional attributes (additionally to those of Preisblatt)



"""
Contains Geraeteeigenschaften and corresponding marshmallow schema for de-/serialization
"""
from typing import Optional

from bo4e.com.com import COM
from bo4e.enum.geraetemerkmal import Geraetemerkmal
from bo4e.enum.geraetetyp import Geraetetyp

# pylint: disable=too-few-public-methods


class Geraeteeigenschaften(COM):
    """
    Mit dieser Komponente werden die Eigenschaften eines Gerätes in Bezug auf den Typ und weitere Merkmale modelliert

    .. raw:: html

        <object data="../_static/images/bo4e/com/Geraeteeigenschaften.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geraeteeigenschaften JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Geraeteeigenschaften.json>`_

    """

    # required attributes
    #: Der Typ eines Gerätes, beispielsweise Drehstromzähler
    geraetetyp: Geraetetyp

    # optional attributes
    #: Weitere Merkmale des Geräts, zum Beispiel Mehrtarif, Eintarif etc..
    geraetemerkmal: Optional[Geraetemerkmal] = None



# pylint: disable=missing-module-docstring, line-too-long
from bo4e.enum.strenum import StrEnum


class Voraussetzungen(StrEnum):
    """
    Voraussetzungen, die erfüllt sein müssen, damit dieser Tarif zur Anwendung kommen kann.
    """

    EINZUGSERMAECHTIGUNG = "EINZUGSERMAECHTIGUNG"  #: Einzugsermaechtigung
    ZEITPUNKT = "ZEITPUNKT"  #: Vertrag muss zu einem bestimmten Zeitpunkt noch bestehen
    LIEFERANBINDUNG_EINE = "LIEFERANBINDUNG_EINE"  #: Lieferantenbindung für diese Energieart
    LIEFERANBINDUNG_ALLE = (
        "LIEFERANBINDUNG_ALLE"  #: Lieferantenbindung für alle Energiearten, die der Versorger anbietet
    )
    GEWERBE = "GEWERBE"  #: Gewerbenachweis
    LASTPROFIL = "LASTPROFIL"  #: Kunde muss einem bestimmten Lastprofil zuzuordnen sein
    ZAEHLERTYP_GROESSE = "ZAEHLERTYP_GROESSE"  #: bestimmter Zaehlertyp oder Zaehlergroeße
    AUSSCHLUSS_GROSSVERBRAUCHER = "AUSSCHLUSS_GROSSVERBRAUCHER"  #: Ausschluss von Großverbrauchern wie Industriekunden oder produzierendes Gewerbe
    NEUKUNDE = "NEUKUNDE"  #: Neukunden ohne bisherige Lieferanbindung
    BESTIMMTE_VERTRAGSFORMALITAETEN = "BESTIMMTE_VERTRAGSFORMALITAETEN"  #: bestimmte Vertragsformalitaeten wie z.B. Anmeldeformular muss an bestimmte Adresse versandt werden
    SELBSTABLESUNG = "SELBSTABLESUNG"  #: Selbstablesung durch den Kunden
    ONLINEVORAUSSETZUNG = "ONLINEVORAUSSETZUNG"  #: Onlinevoraussetzung
    MINDESTUMSATZ = (
        "MINDESTUMSATZ"  #: Mindestumsatz in einer bestimmten Zeiteinheit wie z.B. Mindestjahresumsatz 2500 EURO
    )
    ZUSATZPRODUKT = (
        "ZUSATZPRODUKT"  #: Zusatzprodukt zu bereits bestehendem Lieferverhaeltnis beim Versorger in dieser Energieart
    )
    NEUKUNDE_MIT_VORAUSSETZUNGEN = (
        "NEUKUNDE_MIT_VORAUSSETZUNGEN"  #: geworbener Neukunde muss bestimmte Voraussetzungen erfüllen
    )
    DIREKTVERTRIEB = "DIREKTVERTRIEB"  #: Kunde wird durch Direktvertrieb gewonnen
    ANSCHLUSSART = "ANSCHLUSSART"  #: Anlage mit bestimmter Anschlussart wie z.B. Festanschluss
    ANSCHLUSSWERT = "ANSCHLUSSWERT"  #: bestimmte Anschlusswerte wie z. B. mindestens 30 kW
    ALTER_KUNDENANLAGE = (
        "ALTER_KUNDENANLAGE"  #: Alter einer Kundenanlage (z.B. Anlage wurde nach dem 01.01.2005 installiert)
    )
    ANLAGEBESCHAFFENHEIT = (
        "ANLAGEBESCHAFFENHEIT"  #: bestimmte Anlagebeschaffenheit, wie bivalente Energieversorgung, Geräte o.ä.
    )
    BETRIEBSSTUNDENBEGRENZUNG = (
        "BETRIEBSSTUNDENBEGRENZUNG"  #: Betriebsstundenbegrenzung z.B. max 1500h/a oder mindestens 1000h/a
    )
    FREIGABEZEITEN = "FREIGABEZEITEN"  #: vorgeschriebene Freigabezeiten
    FAMILIENSTRUKTUR = (
        "FAMILIENSTRUKTUR"  #: Familienstruktur wie z.B. bestimmte Anzahl Kinder oder Personen im Haushalt oder Eheleute
    )
    MITGLIEDSCHAFT = "MITGLIEDSCHAFT"  #: Mitgliedschaft in bestimmten Vereinen oder Verbaenden
    STAATLICHE_FOERDERUNG = "STAATLICHE_FOERDERUNG"  #: staatliche Foerderung wie z.B. Sozialtarif oder Studentenausweis
    BESONDERE_VERBRAUCHSSTELLE = (
        "BESONDERE_VERBRAUCHSSTELLE"  #: besondere Verbrauchsstellen wie Kirchen, Vereinsgebaeude usw.
    )
    NIEDRIGENERGIE = "NIEDRIGENERGIE"  #: Niedrigenergieaustattung des Hauses
    ORTSTEILE_LIEFERGEBIET = "ORTSTEILE_LIEFERGEBIET"  #: nur für bestimmte Ortsteile in diesem Liefergebiet
    WAERMEBEDARF_ERDGAS = "WAERMEBEDARF_ERDGAS"  #: Wärmebedarf wird nur oder überwiegend mit Erdgas gedeckt
    MAX_ZAEHLER_LIEFERSTELLEN = "MAX_ZAEHLER_LIEFERSTELLEN"  #: beschränkt auf max. Anzahl Zähler oder Abnahmestellen
    LIEFERUNGSBESCHRAENKUNG_GASART = (
        "LIEFERUNGSBESCHRAENKUNG_GASART"  #: Lieferungsbeschraenkung auf bestimmte Gasart, wie H-Gas oder L-Gas
    )
    KOMBI_BONI = (
        "KOMBI_BONI"  #: Kombination von Boni, von denen mindestens einer sehr unwahrscheinlich zu erreichen ist
    )
    ALTVERTRAG = "ALTVERTRAG"  #: nur für Altvertraege, die weiterhin gueltig sind
    VORGESCHRIEBENE_ZUSATZANLAGE = (
        "VORGESCHRIEBENE_ZUSATZANLAGE"  #: vorgeschriebene Zusatzanlage wie z.B. Solaranlage etc.
    )
    MEHRERE_ZAEHLER_ABNAHMESTELLEN = "MEHRERE_ZAEHLER_ABNAHMESTELLEN"  #: mehr als 1 Zaehler oder 1 Abnahmestelle
    BESTIMMTER_ABNAHMEFALL = (
        "BESTIMMTER_ABNAHMEFALL"  #: bestimmter Abnahmefall wie z.B. nur Gemeinschaftsheizungen o.ae.
    )
    ZUSATZMODALITAET = "ZUSATZMODALITAET"  #: Zahlungsmodalitaet wie z.B. halbjaehrliche Zahlungsweise
    NACHWEIS_ZAHLUNGSFAEHIGKEIT = (
        "NACHWEIS_ZAHLUNGSFAEHIGKEIT"  #: Nachweis der Zahlungsfaehigkeit wie z.B. Bonitaetsprüfung
    )
    UMSTELLUNG_ENERGIEART = "UMSTELLUNG_ENERGIEART"  #: Umstellung der Energieart



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Anrede(StrEnum):
    """
    Übersicht möglicher Anreden, z.B. eines Geschäftspartners.
    """

    HERR = "HERR"  #: "Herr
    FRAU = "FRAU"  #: Frau
    EHELEUTE = "EHELEUTE"  #: Eheleute
    FIRMA = "FIRMA"  #: Firma
    INDIVIDUELL = "INDIVIDUELL"  #: Individuell (z.B. "Profx")



from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.ausschreibungslos import Ausschreibungslos
from bo4e.enum.preismodell import Preismodell
from bo4e.enum.rechnungslegung import Rechnungslegung
from bo4e.enum.sparte import Sparte
from bo4e.enum.vertragsform import Vertragsform
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_ausschreibungsdetail import example_ausschreibungsdetail, example_ausschreibungsdetail_dict
from tests.test_menge import example_menge, example_menge_dict
from tests.test_zeitraum import example_zeitraum, example_zeitraum_dict

example_ausschreibungslos = Ausschreibungslos(
    losnummer="foo",
    bezeichnung="bar",
    bemerkung="asd",
    preismodell=Preismodell.FESTPREIS,
    energieart=Sparte.STROM,
    wunsch_rechnungslegung=Rechnungslegung.MONATSRECHN,
    wunsch_vertragsform=Vertragsform.DIREKT,
    betreut_durch="Max Mustermann",
    anzahl_lieferstellen=17,
    lieferstellen=[example_ausschreibungsdetail],
    gesamt_menge=example_menge,
    wunsch_mindestmenge=example_menge,
    wunsch_maximalmenge=example_menge,
    lieferzeitraum=example_zeitraum,
    wunsch_kuendingungsfrist=example_zeitraum,
    wunsch_zahlungsziel=example_zeitraum,
    wiederholungsintervall=example_zeitraum,
)


class TestAusschreibungslos:
    @pytest.mark.parametrize(
        "ausschreibungslos, expected_json_dict",
        [
            pytest.param(
                example_ausschreibungslos,
                {
                    "lieferzeitraum": example_zeitraum_dict,
                    "preismodell": Preismodell.FESTPREIS,
                    "energieart": Sparte.STROM,
                    "wiederholungsintervall": example_zeitraum_dict,
                    "bemerkung": "asd",
                    "bezeichnung": "bar",
                    "losnummer": "foo",
                    "anzahlLieferstellen": 17,
                    "lieferstellen": [example_ausschreibungsdetail_dict],
                    "wunschKuendingungsfrist": example_zeitraum_dict,
                    "wunschZahlungsziel": example_zeitraum_dict,
                    "gesamtMenge": example_menge_dict,
                    "wunschVertragsform": Vertragsform.DIREKT,
                    "wunschMaximalmenge": example_menge_dict,
                    "wunschRechnungslegung": Rechnungslegung.MONATSRECHN,
                    "wunschMindestmenge": example_menge_dict,
                    "betreutDurch": "Max Mustermann",
                },
                id="maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, ausschreibungslos: Ausschreibungslos, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Ausschreibungslos
        """
        assert_serialization_roundtrip(ausschreibungslos, expected_json_dict)

    def test_ausschreibungslos_lieferstellen_required(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Ausschreibungslos(
                losnummer="foo",
                bezeichnung="bar",
                bemerkung="asd",
                preismodell=Preismodell.FESTPREIS,
                energieart=Sparte.STROM,
                wunsch_rechnungslegung=Rechnungslegung.MONATSRECHN,
                wunsch_vertragsform=Vertragsform.DIREKT,
                betreut_durch="Max Mustermann",
                anzahl_lieferstellen=17,
                lieferzeitraum=example_zeitraum,
                ## ^^ above is just clutter
                lieferstellen=[],  # the important line
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Ausschreibungslos()  # type: ignore[call-arg]
        assert "10 validation errors" in str(excinfo.value)



import pytest
from pydantic import ValidationError

from bo4e.bo.preisblattdienstleistung import PreisblattDienstleistung
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.dienstleistungstyp import Dienstleistungstyp
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.sparte import Sparte
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_geraeteeigenschaften import example_geraeteeigenschaften
from tests.test_marktteilnehmer import example_marktteilnehmer
from tests.test_preisposition import example_preisposition
from tests.test_zeitraum import example_zeitraum


class TestPreisblattDienstleistung:
    @pytest.mark.parametrize(
        "preisblatt_dienstleistung",
        [
            pytest.param(
                PreisblattDienstleistung(
                    bezeichnung="foo",
                    sparte=Sparte.STROM,
                    preisstatus=Preisstatus.ENDGUELTIG,
                    preispositionen=[example_preisposition],
                    gueltigkeit=example_zeitraum,
                    herausgeber=example_marktteilnehmer,
                    bilanzierungsmethode=Bilanzierungsmethode.TLP_GEMEINSAM,
                    basisdienstleistung=Dienstleistungstyp.ABLESUNG_MONATLICH,
                    inklusive_dienstleistungen=[Dienstleistungstyp.AUSLESUNG_FERNAUSLESUNG_ZUSAETZLICH_MSB],
                    geraetedetails=example_geraeteeigenschaften,
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, preisblatt_dienstleistung: PreisblattDienstleistung) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(preisblatt_dienstleistung)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = PreisblattDienstleistung()  # type: ignore[call-arg]
        assert "7 validation errors" in str(excinfo.value)  # 5 from preisblatt + 2 from preisblatt dienstleistung



"""
Contains Vertrag class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from typing import List, Optional

# pylint: disable=unused-argument
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.unterschrift import Unterschrift
from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.com.vertragsteil import Vertragsteil
from bo4e.enum.botyp import BoTyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.vertragsart import Vertragsart
from bo4e.enum.vertragsstatus import Vertragsstatus

# pylint: disable=too-many-instance-attributes, too-few-public-methods


class Vertrag(Geschaeftsobjekt):
    """
    Modell für die Abbildung von Vertragsbeziehungen;
    Das Objekt dient dazu, alle Arten von Verträgen, die in der Energiewirtschaft Verwendung finden, abzubilden.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Vertrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Vertrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Vertrag.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.VERTRAG
    # pylint: disable=duplicate-code
    #: Eine im Verwendungskontext eindeutige Nummer für den Vertrag
    vertragsnummer: str
    #: Hier ist festgelegt, um welche Art von Vertrag es sich handelt.
    vertragsart: Vertragsart
    #: Gibt den Status des Vertrags an
    vertragsstatus: Vertragsstatus
    #: Unterscheidungsmöglichkeiten für die Sparte
    sparte: Sparte
    #: Gibt an, wann der Vertrag beginnt (inklusiv)
    vertragsbeginn: datetime
    #: Gibt an, wann der Vertrag (voraussichtlich) endet oder beendet wurde (exklusiv)
    vertragsende: datetime
    # todo: add von/bis validator
    vertragspartner1: Geschaeftspartner
    """
    Der "erstgenannte" Vertragspartner.
    In der Regel der Aussteller des Vertrags.
    Beispiel: "Vertrag zwischen Vertragspartner 1 ..."
    """
    vertragspartner2: Geschaeftspartner
    """
    Der "zweitgenannte" Vertragspartner.
    In der Regel der Empfänger des Vertrags.
    Beispiel "Vertrag zwischen Vertragspartner 1 und Vertragspartner 2".
    """
    vertragsteile: conlist(Vertragsteil, min_items=1)  # type: ignore[valid-type]
    """
    Der Vertragsteil wird dazu verwendet, eine vertragliche Leistung in Bezug zu einer Lokation
    (Markt- oder Messlokation) festzulegen.
    """

    # optional attributes
    #: Beschreibung zum Vertrag
    beschreibung: Optional[str] = None
    #: Festlegungen zu Laufzeiten und Kündigungsfristen
    vertragskonditionen: Optional[Vertragskonditionen] = None
    #: Unterzeichner des Vertragspartners 1
    unterzeichnervp1: Optional[List[Unterschrift]] = None
    #: Unterzeichner des Vertragspartners 2
    unterzeichnervp2: Optional[List[Unterschrift]] = None



from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.adresse import Adresse
from bo4e.com.externereferenz import ExterneReferenz
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle


class TestExterneReferenz:
    def test_serialization(self) -> None:
        er = ExterneReferenz(ex_ref_name="HOCHFREQUENZ_HFSAP_100", ex_ref_wert="12345")

        er_json = er.json(by_alias=True, ensure_ascii=False)

        assert "exRefName" in er_json

        deserialized_er: ExterneReferenz = ExterneReferenz.parse_raw(er_json)
        assert isinstance(deserialized_er, ExterneReferenz)
        assert deserialized_er == er

    def test_list_of_externe_referenz(self) -> None:
        gp = Geschaeftspartner(
            externe_referenzen=[
                ExterneReferenz(ex_ref_name="SAP GP Nummer", ex_ref_wert="0123456789"),
                ExterneReferenz(ex_ref_name="Schufa-ID", ex_ref_wert="aksdlakoeuhn"),
            ],
            # just some dummy data to make the GP valid
            name1="Duck",
            name2="Donald",
            gewerbekennzeichnung=False,
            geschaeftspartnerrolle=[Geschaeftspartnerrolle.KUNDE],
            partneradresse=Adresse(
                strasse="Am Geldspeicher",
                hausnummer="17",
                postleitzahl="88101",
                ort="Entenhausen",
            ),
        )

        gp_json = gp.json(by_alias=True, ensure_ascii=False)

        deserialized_gp: Geschaeftspartner = Geschaeftspartner.parse_raw(gp_json)
        assert len(deserialized_gp.externe_referenzen) == 2  # type: ignore[arg-type]
        assert deserialized_gp.externe_referenzen[0].ex_ref_name == "SAP GP Nummer"  # type: ignore[index]

    def test_geschaeftspartner_with_no_externe_referenz(self) -> None:
        gp = Geschaeftspartner(
            # just some dummy data to make the GP valid
            name1="Duck",
            name2="Donald",
            gewerbekennzeichnung=False,
            geschaeftspartnerrolle=[Geschaeftspartnerrolle.KUNDE],
            partneradresse=Adresse(
                strasse="Am Geldspeicher",
                hausnummer="17",
                postleitzahl="88101",
                ort="Entenhausen",
            ),
        )

        gp_json = gp.json(by_alias=True, ensure_ascii=False)

        deserialized_gp: Geschaeftspartner = Geschaeftspartner.parse_raw(gp_json)

        assert deserialized_gp.externe_referenzen == []



from bo4e.com.katasteradresse import Katasteradresse


class TestKatasteradresse:
    def test_serialization(self) -> None:
        ka = Katasteradresse(gemarkung_flur="hello", flurstueck="world")

        json_string = ka.json(by_alias=True, ensure_ascii=False)

        assert "gemarkungFlur" in json_string, "No camel case serialization"

        deserialized_ka: Katasteradresse = Katasteradresse.parse_raw(json_string)

        assert ka.gemarkung_flur == deserialized_ka.gemarkung_flur



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Tarifzeit(StrEnum):
    """
    Zur Kennzeichnung verschiedener Tarifzeiten, beispielsweise zur Bepreisung oder zur Verbrauchsermittlung.
    """

    TZ_STANDARD = "TZ_STANDARD"  #: Tarifzeit Standard für Eintarif-Konfigurationen
    TZ_HT = "TZ_HT"  #: Tarifzeit für Hochtarif bei Mehrtarif-Konfigurationen
    TZ_NT = "TZ_NT"  #: Tarifzeit für Niedrigtarif bei Mehrtarif-Konfigurationen



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


# pylint: disable=line-too-long
class Dienstleistungstyp(StrEnum):
    """
    Auflistung möglicher abzurechnender Dienstleistungen.
    """

    DATENBEREITSTELLUNG_TAEGLICH = "DATENBEREITSTELLUNG_TAEGLICH"  #: Datenbereitstellung täglich
    DATENBEREITSTELLUNG_WOECHENTLICH = "DATENBEREITSTELLUNG_WOECHENTLICH"  #: Datenbereitstellung wöchentlich
    DATENBEREITSTELLUNG_MONATLICH = "DATENBEREITSTELLUNG_MONATLICH"  #: Datenbereitstellung monatlich
    DATENBEREITSTELLUNG_JAEHRLICH = "DATENBEREITSTELLUNG_JAEHRLICH"  #: Datenbereitstellung jährlich
    DATENBEREITSTELLUNG_HISTORISCHE_LG = "DATENBEREITSTELLUNG_HISTORISCHE_LG"
    # Datenbereitstellung historischer Lastgänge
    DATENBEREITSTELLUNG_STUENDLICH = "DATENBEREITSTELLUNG_STUENDLICH"  #: Datenbereitstellung stündlich
    DATENBEREITSTELLUNG_VIERTELJAEHRLICH = "DATENBEREITSTELLUNG_VIERTELJAEHRLICH"
    # Datenbereitstellung vierteljährlich
    DATENBEREITSTELLUNG_HALBJAEHRLICH = "DATENBEREITSTELLUNG_HALBJAEHRLICH"  #: Datenbereitstellung halbjährlich
    DATENBEREITSTELLUNG_MONATLICH_ZUSAETZLICH = "DATENBEREITSTELLUNG_MONATLICH_ZUSAETZLICH"
    # Datenbereitstellung monatlich zusätzlich
    DATENBEREITSTELLUNG_EINMALIG = "DATENBEREITSTELLUNG_EINMALIG"  #: Datenbereitstellung einmalig
    AUSLESUNG_2X_TAEGLICH_FERNAUSLESUNG = "AUSLESUNG_2X_TAEGLICH_FERNAUSLESUNG"
    # Auslesung 2x täglich mittels Fernauslesung
    AUSLESUNG_TAEGLICH_FERNAUSLESUNG = "AUSLESUNG_TAEGLICH_FERNAUSLESUNG"  #: Auslesung täglich mittels Fernauslesung
    AUSLESUNG_MANUELL_MSB = "AUSLESUNG_MANUELL_MSB"  #: Auslesung, manuell vom Messstellenbetreiber vorgenommen
    AUSLESUNG_MONATLICH_FERNAUSLESUNG = "AUSLESUNG_MONATLICH_FERNAUSLESUNG"
    # Auslesung monatlich bei mittels Fernauslesung
    AUSLESUNG_JAEHRLICH_FERNAUSLESUNG = "AUSLESUNG_JAEHRLICH_FERNAUSLESUNG"
    # Auslesung jährlich bei SLP mittels Fernauslesung
    AUSLESUNG_MDE = "AUSLESUNG_MDE"  #: Auslesung mit mobiler Daten Erfassung (MDE)
    ABLESUNG_MONATLICH = "ABLESUNG_MONATLICH"  #: Ablesung monatlich
    ABLESUNG_VIERTELJAEHRLICH = "ABLESUNG_VIERTELJAEHRLICH"  #: Ablesung vierteljährlich
    ABLESUNG_HALBJAEHRLICH = "ABLESUNG_HALBJAEHRLICH"  #: Ablesung halbjährlich
    ABLESUNG_JAEHRLICH = "ABLESUNG_JAEHRLICH"  #: Ablesung jährlich
    AUSLESUNG_FERNAUSLESUNG = "AUSLESUNG_FERNAUSLESUNG"  #: Auslesung mittels Fernauslesung
    ABLESUNG_ZUSAETZLICH_MSB = "ABLESUNG_ZUSAETZLICH_MSB"  #: Ablesung, zusätzlich vom Messstellenbetreiber vorgenommen
    ABLESUNG_ZUSAETZLICH_KUNDE = "ABLESUNG_ZUSAETZLICH_KUNDE"  #: Ablesung SLP, zusätzlich vom Kunden vorgenommen
    AUSLESUNG_FERNAUSLESUNG_ZUSAETZLICH_MSB = "AUSLESUNG_FERNAUSLESUNG_ZUSAETZLICH_MSB"
    # Auslesung, mittels Fernauslesung, zusätzlich vom Messstellenbetreiber vorgenommen
    AUSLESUNG_MOATLICH_FERNAUSLESUNG = "AUSLESUNG_MOATLICH_FERNAUSLESUNG"  #: Auslesung monatlich mittels Fernauslesung
    AUSLESUNG_STUENDLICH_FERNAUSLESUNG = "AUSLESUNG_STUENDLICH_FERNAUSLESUNG"
    # Auslesung stündlich mittels Fernauslesung
    AUSLESUNG_TEMPERATURMENGENUMWERTER = "AUSLESUNG_TEMPERATURMENGENUMWERTER"  #: Auslesung Temperaturmengenumwerter
    AUSLESUNG_ZUSTANDSMENGENUMWERTER = "AUSLESUNG_ZUSTANDSMENGENUMWERTER"  #: Auslesung Zustandsmengenumwerter
    AUSLESUNG_SYSTEMMENGENUMWERTER = "AUSLESUNG_SYSTEMMENGENUMWERTER"  #: Auslesung Systemmengenumwerter
    AUSLESUNG_VORGANG = "AUSLESUNG_VORGANG"  #: Auslesung je Vorgang
    AUSLESUNG_KOMPAKTMENGENUMWERTER = "AUSLESUNG_KOMPAKTMENGENUMWERTER"  #: Auslesung Kompaktmengenumwerter
    SPERRUNG = "SPERRUNG"  #: Sperrung einer Abnahmestelle
    ENTSPERRUNG = "ENTSPERRUNG"  #: Entsperrung einer Abnahmestelle
    MAHNKOSTEN = "MAHNKOSTEN"  #: Mahnkosten
    INKASSOKOSTEN = "INKASSOKOSTEN"  #: Inkassokosten



from decimal import Decimal

import pytest

from bo4e.com.fremdkostenposition import Fremdkostenposition
from bo4e.com.tarifberechnungsparameter import Tarifberechnungsparameter
from bo4e.enum.messpreistyp import Messpreistyp
from bo4e.enum.tarifkalkulationsmethode import Tarifkalkulationsmethode
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_preis import example_preis
from tests.test_tarifpreis import example_tarifpreis

example_tarifberechnungsparameter = Tarifberechnungsparameter(
    berechnungsmethode=Tarifkalkulationsmethode.ZONEN,
    messpreis_in_gp_enthalten=True,
    kw_inklusive=Decimal(12.5),
    kw_weitere_mengen=Decimal(12.5),
    messpreistyp=Messpreistyp.MESSPREIS_G6,
    messpreis_beruecksichtigen=True,
    hoechstpreis_h_t=example_preis,
    hoechstpreis_n_t=example_preis,
    mindestpreis=example_preis,
    zusatzpreise=[example_tarifpreis],
)


class TestFremdkostenposition:
    @pytest.mark.parametrize(
        "tarifberechnungsparameter",
        [
            pytest.param(
                example_tarifberechnungsparameter,
                id="maximal attributes",
            ),
            pytest.param(
                Tarifberechnungsparameter(),
                id="minimal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifberechnungsparameter: Fremdkostenposition) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarifberechnungsparameter)

    def test_missing_required_attribute(self) -> None:
        _ = Tarifberechnungsparameter()
        # ok, we're done. no exception here because there are no required attributes



from datetime import datetime, timezone
from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.menge import Menge
from bo4e.com.vertragsteil import Vertragsteil
from bo4e.enum.mengeneinheit import Mengeneinheit


class TestVertragsteil:
    def test_vertragsteil_only_required_attributes(self) -> None:
        """
        Test de-/serialisation of Vertragsteil with minimal attributes.
        """
        vertragsteil = Vertragsteil(
            vertragsteilbeginn=datetime(2001, 3, 15, tzinfo=timezone.utc),
            vertragsteilende=datetime(2007, 11, 27, tzinfo=timezone.utc),
        )

        json_string = vertragsteil.json(by_alias=True, ensure_ascii=False)

        assert "2001-03-15T00:00:00+00:00" in json_string
        assert "2007-11-27T00:00:00+00:00" in json_string

        vertragsteil_deserialized = Vertragsteil.parse_raw(json_string)

        assert isinstance(vertragsteil_deserialized.vertragsteilbeginn, datetime)
        assert vertragsteil_deserialized.vertragsteilbeginn == datetime(2001, 3, 15, tzinfo=timezone.utc)
        assert isinstance(vertragsteil_deserialized.vertragsteilende, datetime)
        assert vertragsteil_deserialized.vertragsteilende == datetime(2007, 11, 27, tzinfo=timezone.utc)

    def test_vertragsteil_required_and_optional_attributes(self) -> None:
        """
        Test de-/serialisation of Vertragsteil with maximal attributes.
        """
        vertragsteil = Vertragsteil(
            vertragsteilbeginn=datetime(2001, 3, 15, tzinfo=timezone.utc),
            vertragsteilende=datetime(2007, 11, 27, tzinfo=timezone.utc),
            lokation="Bar",
            vertraglich_fixierte_menge=Menge(wert=Decimal(3.1), einheit=Mengeneinheit.KWH),
            minimale_abnahmemenge=Menge(wert=Decimal(2000), einheit=Mengeneinheit.KWH),
            maximale_abnahmemenge=Menge(wert=Decimal(0.111111), einheit=Mengeneinheit.KWH),
        )

        json_string = vertragsteil.json(by_alias=True, ensure_ascii=False)

        assert "2001-03-15T00:00:00+00:00" in json_string
        assert "2007-11-27T00:00:00+00:00" in json_string
        assert "Bar" in json_string
        assert "KWH" in json_string
        assert "0.111111" in json_string

        vertragsteil_deserialized = Vertragsteil.parse_raw(json_string)

        assert isinstance(vertragsteil_deserialized.vertragsteilbeginn, datetime)
        assert vertragsteil_deserialized.vertragsteilbeginn == datetime(2001, 3, 15, tzinfo=timezone.utc)
        assert isinstance(vertragsteil_deserialized.vertragsteilende, datetime)
        assert vertragsteil_deserialized.vertragsteilende == datetime(2007, 11, 27, tzinfo=timezone.utc)
        assert isinstance(vertragsteil_deserialized.lokation, str)
        assert vertragsteil_deserialized.lokation == "Bar"
        assert isinstance(vertragsteil_deserialized.minimale_abnahmemenge, Menge)
        assert vertragsteil_deserialized.minimale_abnahmemenge == Menge(wert=Decimal(2000), einheit=Mengeneinheit.KWH)

    def test_vertragsteil_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Vertragsteil(vertragsteilende=datetime(2007, 11, 27, tzinfo=timezone.utc))  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)



from datetime import datetime, timezone
from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.zeitreihenwert import Zeitreihenwert
from bo4e.enum.messwertstatus import Messwertstatus
from bo4e.enum.messwertstatuszusatz import Messwertstatuszusatz

example_zeitreihenwert = Zeitreihenwert(
    wert=Decimal(2.5),
    datum_uhrzeit_von=datetime(2001, 3, 15, tzinfo=timezone.utc),
    datum_uhrzeit_bis=datetime(2007, 11, 27, tzinfo=timezone.utc),
)


class TestZeitreihenwert:
    def test_zeitreihenwert_only_required_attributes(self) -> None:
        """
        Test de-/serialisation of Zeitreihenwert with minimal attributes.
        """
        zeitreihenwert = example_zeitreihenwert

        json_string = zeitreihenwert.json(by_alias=True, ensure_ascii=False)

        assert "2001-03-15T00:00:00+00:00" in json_string
        assert "2007-11-27T00:00:00+00:00" in json_string

        zeitreihenwert_deserialized: Zeitreihenwert = Zeitreihenwert.parse_raw(json_string)
        assert zeitreihenwert_deserialized == zeitreihenwert

    def test_zeitreihenwert_required_and_optional_attributes(self) -> None:
        """
        Test de-/serialisation of Zeitreihenwert with maximal attributes.
        """
        zeitreihenwert = Zeitreihenwert(
            datum_uhrzeit_von=datetime(2001, 3, 15, tzinfo=timezone.utc),
            datum_uhrzeit_bis=datetime(2007, 11, 27, tzinfo=timezone.utc),
            wert=Decimal(2.5),
            status=Messwertstatus.ABGELESEN,
            statuszusatz=Messwertstatuszusatz.Z78_GERAETEWECHSEL,
        )

        json_string = zeitreihenwert.json(by_alias=True, ensure_ascii=False)

        assert "2.5" in json_string
        assert "2001-03-15T00:00:00+00:00" in json_string
        assert "2007-11-27T00:00:00+00:00" in json_string
        assert "ABGELESEN" in json_string
        assert "Z78_GERAETEWECHSEL" in json_string

        zeitreihenwert_deserialized: Zeitreihenwert = Zeitreihenwert.parse_raw(json_string)
        assert zeitreihenwert_deserialized == zeitreihenwert

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitreihenwert(datum_uhrzeit_von=datetime(2007, 11, 27, tzinfo=timezone.utc), wert=Decimal(1.5))  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)

    def test_von_later_than_bis(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitreihenwert(
                datum_uhrzeit_von=datetime(2007, 11, 27, tzinfo=timezone.utc),
                datum_uhrzeit_bis=datetime(2006, 11, 27, tzinfo=timezone.utc),
                wert=Decimal(1.5),
            )

        assert "start" in str(excinfo.value)
        assert "end" in str(excinfo.value)



"""
Contains Preis class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal
from typing import Optional

from bo4e.com.com import COM
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.waehrungseinheit import Waehrungseinheit

# pylint: disable=too-few-public-methods


class Preis(COM):
    """
    Abbildung eines Preises mit Wert, Einheit, Bezugswert und Status.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Preis.svg" type="image/svg+xml"></object>

    .. HINT::
        `Preis JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Preis.json>`_

    """

    # required attributes
    #: Gibt die nominale Höhe des Preises an.
    wert: Decimal
    #: Währungseinheit für den Preis, z.B. Euro oder Ct.
    einheit: Waehrungseinheit
    #: Angabe, für welche Bezugsgröße der Preis gilt. Z.B. kWh.
    bezugswert: Mengeneinheit

    # optional attributes
    #: Gibt den Status des veröffentlichten Preises an
    status: Optional[Preisstatus] = None



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.geokoordinaten import Geokoordinaten


class TestGeokoordinaten:
    def test_serialization(self) -> None:
        geo = Geokoordinaten(
            breitengrad=Decimal(52.52149200439453),
            laengengrad=Decimal(13.404866218566895),
        )

        json_string = geo.json(by_alias=True, ensure_ascii=False)

        assert "breitengrad" in json_string
        assert str(geo.breitengrad) in json_string

        deserialized_geo: Geokoordinaten = Geokoordinaten.parse_raw(json_string)

        assert isinstance(deserialized_geo.breitengrad, Decimal)
        assert isinstance(deserialized_geo.laengengrad, Decimal)
        assert geo.breitengrad == deserialized_geo.breitengrad

    def test_wrong_datatype(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Geokoordinaten(breitengrad="54,23", laengengrad=-23.2)  # type: ignore[arg-type]

        assert "breitengrad" in str(excinfo.value)



# pylint: disable=missing-module-docstring
from decimal import Decimal
from typing import List, Optional

from humps.main import camelize

# pylint: disable=no-name-in-module
from pydantic import BaseModel

from bo4e.com.externereferenz import ExterneReferenz
from bo4e.enum.botyp import BoTyp

# pylint: disable=too-few-public-methods


class Geschaeftsobjekt(BaseModel):
    """
    Das BO Geschäftsobjekt ist der Master für alle Geschäftsobjekte.
    Alle Attribute, die hier in diesem BO enthalten sind, werden an alle BOs vererbt.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Geschaeftsobjekt.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geschaeftsobjekt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Geschaeftsobjekt.json>`_

    """

    # required attributes
    versionstruktur: str = "2"  #: Version der BO-Struktur aka "fachliche Versionierung"
    bo_typ: BoTyp = BoTyp.GESCHAEFTSOBJEKT  #: Der Typ des Geschäftsobjektes
    # bo_typ is used as discriminator f.e. for databases or deserialization

    # optional attributes
    externe_referenzen: Optional[List[ExterneReferenz]] = []
    #: Hier können IDs anderer Systeme hinterlegt werden (z.B. eine SAP-GP-Nummer oder eine GUID)

    class Config:
        """
        basic configuration for pydantic's behaviour
        """

        alias_generator = camelize
        allow_population_by_field_name = True
        json_encoders = {Decimal: str}



"""
Contains Tarifpreisposition class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.com.com import COM
from bo4e.com.preisstaffel import Preisstaffel
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preistyp import Preistyp
from bo4e.enum.waehrungseinheit import Waehrungseinheit


class Tarifpreisposition(COM):
    """
    Mit dieser Komponente können Tarifpreise verschiedener Typen abgebildet werden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Tarifpreisposition.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifpreisposition JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Tarifpreisposition.json>`_

    """

    # required attributes
    #: Angabe des Preistypes (z.B. Grundpreis)
    preistyp: Preistyp
    #: Einheit des Preises (z.B. EURO)
    einheit: Waehrungseinheit
    #: Größe, auf die sich die Einheit bezieht, beispielsweise kWh, Jahr
    bezugseinheit: Mengeneinheit
    #: Hier sind die Staffeln mit ihren Preisenangaben definiert
    preisstaffeln: conlist(Preisstaffel, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Gibt an, nach welcher Menge die vorgenannte Einschränkung erfolgt (z.B. Jahresstromverbrauch in kWh)
    mengeneinheitstaffel: Optional[Mengeneinheit] = None



"""
Contains PreisblattKonzessionsabgabe class and corresponding marshmallow schema for de-/serialization
"""


from bo4e.bo.preisblatt import Preisblatt
from bo4e.enum.botyp import BoTyp
from bo4e.enum.kundengruppeka import KundengruppeKA

# pylint: disable=too-few-public-methods


class PreisblattKonzessionsabgabe(Preisblatt):
    """
    Die Variante des Preisblattmodells zur Abbildung von allgemeinen Abgaben

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattKonzessionsabgabe.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattKonzessionsabgabe JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/PreisblattKonzessionsabgabe.json>`_

    """

    bo_typ: BoTyp = BoTyp.PREISBLATTKONZESSIONSABGABE
    # required attributes (additional to those of Preisblatt)
    #: Kundegruppe anhand derer die Höhe der Konzessionabgabe festgelegt ist
    kundengruppe_k_a: KundengruppeKA

    # there are no optional attributes (additionally to those of Preisblatt)



import pytest
from pydantic import ValidationError

from bo4e.bo.lastgang import Lastgang
from bo4e.enum.lokationstyp import Lokationstyp
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.sparte import Sparte
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_zeitreihenwert import example_zeitreihenwert


class TestLastgang:
    @pytest.mark.parametrize(
        "lastgang_kompakt",
        [
            pytest.param(
                Lastgang(
                    version="1.1",
                    sparte=Sparte.STROM,
                    lokations_id="DE0000011111222223333344444555556",
                    obis_kennzahl="1-0:1.8.1",
                    lokationstyp=Lokationstyp.MELO,
                    messgroesse=Mengeneinheit.KWH,
                    werte=[example_zeitreihenwert],
                ),
            ),
        ],
    )
    def test_serialization_roundtrip(self, lastgang_kompakt: Lastgang) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(lastgang_kompakt)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Lastgang()  # type: ignore[call-arg]

        assert "5 validation errors" in str(excinfo.value)



from datetime import datetime, timezone
from decimal import Decimal

from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.zeiteinheit import Zeiteinheit

example_vertragskonditionen = Vertragskonditionen(
    beschreibung="Foobar",
    anzahl_abschlaege=Decimal(3),
    vertragslaufzeit=Zeitraum(
        startdatum=datetime(2012, 9, 21, tzinfo=timezone.utc),
        enddatum=datetime(2013, 10, 11, tzinfo=timezone.utc),
    ),
    kuendigungsfrist=Zeitraum(einheit=Zeiteinheit.WOCHE, dauer=Decimal(3)),
    vertragsverlaengerung=Zeitraum(einheit=Zeiteinheit.TAG, dauer=Decimal(14)),
    abschlagszyklus=Zeitraum(einheit=Zeiteinheit.TAG, dauer=Decimal(5)),
)


class TestVertragskonditionen:
    def test_vertragskonditionen_with_optional_attributes(self) -> None:
        """
        Test de-/serialisation of Vertragskonditionen (only has optional attributes).
        """
        vertragskonditionen = example_vertragskonditionen

        json_string = vertragskonditionen.json(by_alias=True, ensure_ascii=False)

        assert "Foobar" in json_string
        assert "3" in json_string
        assert "2013-10-11T00:00:00+00:00" in json_string
        assert "WOCHE" in json_string
        assert "TAG" in json_string
        assert "14" in json_string

        vertragskonditionen_deserialized = Vertragskonditionen.parse_raw(json_string)

        assert isinstance(vertragskonditionen_deserialized.beschreibung, str)
        assert vertragskonditionen_deserialized.beschreibung == "Foobar"
        assert isinstance(vertragskonditionen_deserialized.anzahl_abschlaege, Decimal)
        assert vertragskonditionen_deserialized.anzahl_abschlaege == Decimal(3)
        assert isinstance(vertragskonditionen_deserialized.vertragslaufzeit, Zeitraum)
        assert vertragskonditionen_deserialized.vertragslaufzeit == Zeitraum(
            startdatum=datetime(2012, 9, 21, tzinfo=timezone.utc), enddatum=datetime(2013, 10, 11, tzinfo=timezone.utc)
        )
        assert isinstance(vertragskonditionen_deserialized.vertragsverlaengerung, Zeitraum)
        assert vertragskonditionen_deserialized.vertragsverlaengerung == Zeitraum(
            einheit=Zeiteinheit.TAG, dauer=Decimal(14)
        )

    def test_vertragskonditionen_empty(self) -> None:
        """
        Test empty Vertragskonditionen (only has optional attributes).
        """
        empty_vertragskonditionen = Vertragskonditionen()

        assert isinstance(empty_vertragskonditionen, Vertragskonditionen)



"""
Contains Angebotsteil class
and corresponding marshmallow schema for de-/serialization
"""

from typing import List, Optional

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.bo.marktlokation import Marktlokation
from bo4e.com.angebotsposition import Angebotsposition
from bo4e.com.betrag import Betrag
from bo4e.com.com import COM
from bo4e.com.menge import Menge
from bo4e.com.zeitraum import Zeitraum


class Angebotsteil(COM):
    """
    Mit dieser Komponente wird ein Teil einer Angebotsvariante abgebildet.
    Hier werden alle Angebotspositionen aggregiert.
    Angebotsteile werden im einfachsten Fall für eine Marktlokation oder Lieferstellenadresse erzeugt.
    Hier werden die Mengen und Gesamtkosten aller Angebotspositionen zusammengefasst.
    Eine Variante besteht mindestens aus einem Angebotsteil.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Angebotsteil.svg" type="image/svg+xml"></object>

    .. HINT::
        `Angebotsteil JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Angebotsteil.json>`_

    """

    # required attributes
    #: Einzelne Positionen, die zu diesem Angebotsteil gehören
    positionen: conlist(Angebotsposition, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Identifizierung eines Subkapitels einer Anfrage, beispielsweise das Los einer Ausschreibung
    anfrage_subreferenz: Optional[str] = None
    lieferstellenangebotsteil: Optional[List[Marktlokation]] = None
    """
    Marktlokationen, für die dieses Angebotsteil gilt, falls vorhanden.
    Durch die Marktlokation ist auch die Lieferadresse festgelegt
    """
    #: Summe der Verbräuche aller in diesem Angebotsteil eingeschlossenen Lieferstellen
    gesamtmengeangebotsteil: Optional[Menge] = None
    #: Summe der Jahresenergiekosten aller in diesem Angebotsteil enthaltenen Lieferstellen
    gesamtkostenangebotsteil: Optional[Betrag] = None
    #: Hier kann der Belieferungszeitraum angegeben werden, für den dieser Angebotsteil gilt
    lieferzeitraum: Optional[Zeitraum] = None



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.preis import Preis
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.waehrungseinheit import Waehrungseinheit

example_preis = Preis(wert=Decimal(12.5), einheit=Waehrungseinheit.EUR, bezugswert=Mengeneinheit.KWH)


class TestPreis:
    def test_preis_only_required(self) -> None:
        """
        Test de-/serialisation of Preis (only has required attributes).
        """
        preis = example_preis

        json_string = preis.json(by_alias=True, ensure_ascii=False)

        assert "KWH" in json_string
        assert "EUR" in json_string
        assert "null" in json_string

        preis_deserialized = Preis.parse_raw(json_string)

        assert isinstance(preis_deserialized.wert, Decimal)
        assert isinstance(preis_deserialized.einheit, Waehrungseinheit)
        assert isinstance(preis_deserialized.bezugswert, Mengeneinheit)
        assert preis_deserialized.status is None
        assert preis == preis_deserialized

    def test_wrong_datatype(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Preis(wert="lululululu", einheit=Waehrungseinheit.EUR, bezugswert=Mengeneinheit.KWH)  # type: ignore[arg-type]

        assert "1 validation error" in str(excinfo.value)
        assert "wert" in str(excinfo.value)
        assert "value is not a valid decimal" in str(excinfo.value)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Preis(wert=Decimal(3.50), einheit=Waehrungseinheit.EUR, status=Preisstatus.ENDGUELTIG)  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)

    def test_optional_attribute(self) -> None:
        preis = Preis(
            wert=Decimal(3.50),
            einheit=Waehrungseinheit.EUR,
            bezugswert=Mengeneinheit.KWH,
            status=Preisstatus.ENDGUELTIG,
        )

        json_string = preis.json(by_alias=True, ensure_ascii=False)

        assert "ENDGUELTIG" in json_string

        preis_deserialized = Preis.parse_raw(json_string)

        assert isinstance(preis_deserialized.status, Preisstatus)



"""
Contains KriteriumWert class
and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods
from bo4e.enum.tarifregionskriterium import Tarifregionskriterium


class KriteriumWert(COM):
    """
    Mit dieser Komponente können Kriterien und deren Werte definiert werden

    .. raw:: html

        <object data="../_static/images/bo4e/com/KriteriumWert.svg" type="image/svg+xml"></object>

    .. HINT::
        `KriteriumWert JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/KriteriumWert.json>`_

    """

    # required attributes
    #: Hier steht, für welches Kriterium der Wert gilt. Z.B. Postleitzahlen
    kriterium: Tarifregionskriterium
    #: Ein Wert, passend zum Kriterium. Z.B. eine Postleitzahl.
    wert: str



from datetime import datetime, timezone
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.bo.buendelvertrag import Buendelvertrag
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.adresse import Adresse
from bo4e.com.unterschrift import Unterschrift
from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.kontaktart import Kontaktart
from bo4e.enum.landescode import Landescode
from bo4e.enum.sparte import Sparte
from bo4e.enum.vertragsart import Vertragsart
from bo4e.enum.vertragsstatus import Vertragsstatus
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_vertrag import TestVertrag


class TestBuendelvertrag:
    _vertragspartner1 = Geschaeftspartner(
        anrede=Anrede.FRAU,
        name1="van der Waal",
        name2="Helga",
        name3=None,
        gewerbekennzeichnung=True,
        kontaktweg=[Kontaktart.SMS],
        umsatzsteuer_id="DE267311963",
        glaeubiger_id="DE98ZZZ09999999999",
        e_mail_adresse="helga.waal@bo4e.de",
        website="bo4e.de",
        geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
        partneradresse=Adresse(
            postleitzahl="33333",
            ort="Quadrat-Ichendorf",
            strasse="Kubikstraße",
            hausnummer="4",
        ),
    )
    _vertragspartner2 = Geschaeftspartner(
        name1="Eckart",
        name2="Björn",
        gewerbekennzeichnung=False,
        geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
        partneradresse=Adresse(
            postleitzahl="24211",
            ort="Preetz",
            strasse="Am Markt",
            hausnummer="67",
        ),
    )
    _vertragspartner1_dict: Dict[str, Any] = {
        "anrede": Anrede.FRAU,
        "name1": "van der Waal",
        "name2": "Helga",
        "name3": None,
        "gewerbekennzeichnung": True,
        "kontaktweg": [Kontaktart.SMS],
        "umsatzsteuerId": "DE267311963",
        "glaeubigerId": "DE98ZZZ09999999999",
        "eMailAdresse": "helga.waal@bo4e.de",
        "website": "bo4e.de",
        "geschaeftspartnerrolle": [Geschaeftspartnerrolle.DIENSTLEISTER],
        "partneradresse": {
            "postleitzahl": "33333",
            "ort": "Quadrat-Ichendorf",
            "strasse": "Kubikstraße",
            "hausnummer": "4",
            "postfach": None,
            "adresszusatz": None,
            "coErgaenzung": None,
            "landescode": Landescode.DE,  # type:ignore[attr-defined]
        },
        "versionstruktur": "2",
        "boTyp": BoTyp.GESCHAEFTSPARTNER,
        "externeReferenzen": [],
        "hrnummer": None,
        "amtsgericht": None,
    }
    _vertragspartner2_dict: Dict[str, Any] = {
        "name1": "Eckart",
        "name2": "Björn",
        "gewerbekennzeichnung": False,
        "geschaeftspartnerrolle": [Geschaeftspartnerrolle.DIENSTLEISTER],
        "partneradresse": {
            "postleitzahl": "24211",
            "ort": "Preetz",
            "strasse": "Am Markt",
            "hausnummer": "67",
            "postfach": None,
            "adresszusatz": None,
            "coErgaenzung": None,
            "landescode": Landescode.DE,  # type:ignore[attr-defined]
        },
        "versionstruktur": "2",
        "boTyp": BoTyp.GESCHAEFTSPARTNER,
        "externeReferenzen": [],
        "anrede": None,
        "name3": None,
        "hrnummer": None,
        "amtsgericht": None,
        "kontaktweg": [],
        "umsatzsteuerId": None,
        "glaeubigerId": None,
        "eMailAdresse": None,
        "website": None,
    }

    @pytest.mark.parametrize(
        "buendelvertrag, expected_dict",
        [
            pytest.param(
                Buendelvertrag(
                    vertragsnummer="1234567890",
                    vertragsart=Vertragsart.NETZNUTZUNGSVERTRAG,
                    vertragsstatus=Vertragsstatus.AKTIV,
                    sparte=Sparte.STROM,
                    vertragsbeginn=datetime(2021, 4, 30, 13, 45, tzinfo=timezone.utc),
                    vertragsende=datetime(2200, 4, 30, 13, 45, tzinfo=timezone.utc),
                    vertragspartner1=_vertragspartner1,
                    vertragspartner2=_vertragspartner2,
                ),
                {
                    "vertragsnummer": "1234567890",
                    "vertragsart": Vertragsart.NETZNUTZUNGSVERTRAG,
                    "vertragsstatus": Vertragsstatus.AKTIV,
                    "sparte": Sparte.STROM,
                    "vertragsbeginn": datetime(2021, 4, 30, 13, 45, tzinfo=timezone.utc),
                    "vertragsende": datetime(2200, 4, 30, 13, 45, tzinfo=timezone.utc),
                    "vertragspartner1": _vertragspartner1_dict,
                    "vertragspartner2": _vertragspartner2_dict,
                    "versionstruktur": "2",
                    "boTyp": BoTyp.BUENDELVERTRAG,
                    "externeReferenzen": [],
                    "einzelvertraege": [],
                    "vertragskonditionen": [],
                    "unterzeichnervp1": [],
                    "unterzeichnervp2": [],
                    "beschreibung": None,
                },
                id="minimal fields",
            ),
            pytest.param(
                Buendelvertrag(
                    vertragsnummer="1234567890",
                    vertragsart=Vertragsart.NETZNUTZUNGSVERTRAG,
                    vertragsstatus=Vertragsstatus.AKTIV,
                    sparte=Sparte.STROM,
                    vertragsbeginn=datetime(2021, 4, 30, 13, 45, tzinfo=timezone.utc),
                    vertragsende=datetime(2200, 4, 30, 13, 45, tzinfo=timezone.utc),
                    vertragspartner1=_vertragspartner1,
                    vertragspartner2=_vertragspartner2,
                    einzelvertraege=[TestVertrag().get_example_vertrag()],
                    vertragskonditionen=[Vertragskonditionen(beschreibung="Hello World")],
                    unterzeichnervp1=[Unterschrift(name="Helga van der Waal")],
                    unterzeichnervp2=[Unterschrift(name="Björn oder so"), Unterschrift(name="Zweiter Typ")],
                    beschreibung="Das ist ein Bündelvertrag mit allen optionalen Feldern ausgefüllt.",
                ),
                {
                    "versionstruktur": "2",
                    "boTyp": BoTyp.BUENDELVERTRAG,
                    "externeReferenzen": [],
                    "vertragsnummer": "1234567890",
                    "vertragsart": Vertragsart.NETZNUTZUNGSVERTRAG,
                    "vertragsstatus": Vertragsstatus.AKTIV,
                    "sparte": Sparte.STROM,
                    "vertragsbeginn": datetime(2021, 4, 30, 13, 45, tzinfo=timezone.utc),
                    "vertragsende": datetime(2200, 4, 30, 13, 45, tzinfo=timezone.utc),
                    "vertragspartner1": _vertragspartner1_dict,
                    "vertragspartner2": _vertragspartner2_dict,
                    "einzelvertraege": [TestVertrag().get_example_vertrag_expected_dict()],
                    "vertragskonditionen": [
                        {
                            "beschreibung": "Hello World",
                            "anzahlAbschlaege": None,
                            "vertragslaufzeit": None,
                            "kuendigungsfrist": None,
                            "vertragsverlaengerung": None,
                            "abschlagszyklus": None,
                        }
                    ],
                    "unterzeichnervp1": [{"name": "Helga van der Waal", "ort": None, "datum": None}],
                    "unterzeichnervp2": [
                        {"name": "Björn oder so", "ort": None, "datum": None},
                        {"name": "Zweiter Typ", "ort": None, "datum": None},
                    ],
                    "beschreibung": "Das ist ein Bündelvertrag mit allen optionalen Feldern ausgefüllt.",
                },
                id="maximal fields",
            ),
        ],
    )
    def test_serialization_roundtrip(self, buendelvertrag: Buendelvertrag, expected_dict: Dict[str, Any]) -> None:
        assert_serialization_roundtrip(buendelvertrag, expected_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Buendelvertrag()  # type: ignore[call-arg]

        assert "8 validation errors" in str(excinfo.value)



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.steuerbetrag import Steuerbetrag
from bo4e.enum.steuerkennzeichen import Steuerkennzeichen
from bo4e.enum.waehrungscode import Waehrungscode
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip

example_steuerbetrag = Steuerbetrag(
    steuerkennzeichen=Steuerkennzeichen.UST_7,
    basiswert=Decimal(100),
    steuerwert=Decimal(19),
    waehrung=Waehrungscode.EUR,
)


class TestSteuerbetrag:
    @pytest.mark.parametrize(
        "steuerbetrag, expected_json_dict",
        [
            pytest.param(
                example_steuerbetrag,
                {
                    "steuerkennzeichen": "UST_7",
                    "basiswert": Decimal("100"),
                    "steuerwert": Decimal("19"),
                    "waehrung": Waehrungseinheit.EUR,
                },
            ),
        ],
    )
    def test_steuerbetrag_required_attributes(
        self, steuerbetrag: Steuerbetrag, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Steuerbetrag with minimal attributes.
        """
        assert_serialization_roundtrip(steuerbetrag, expected_json_dict)

    def test_steuerbetrag_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Steuerbetrag()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)



import pytest
from pydantic import ValidationError

from bo4e.bo.standorteigenschaften import Standorteigenschaften
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_standorteigenschaftengas import example_standorteigenschaften_gas
from tests.test_standorteigenschaftenstrom import example_standorteigenschaften_strom


class TestStandorteigenschaften:
    @pytest.mark.parametrize(
        "standorteigenschaften",
        [
            pytest.param(
                Standorteigenschaften(
                    eigenschaften_strom=[example_standorteigenschaften_strom],
                    eigenschaften_gas=example_standorteigenschaften_gas,
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, standorteigenschaften: Standorteigenschaften) -> None:
        assert_serialization_roundtrip(standorteigenschaften)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Standorteigenschaften()  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)



from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.geraeteeigenschaften import Geraeteeigenschaften
from bo4e.enum.geraetemerkmal import Geraetemerkmal
from bo4e.enum.geraetetyp import Geraetetyp
from tests.serialization_helper import assert_serialization_roundtrip

example_geraeteeigenschaften = Geraeteeigenschaften(
    geraetemerkmal=Geraetemerkmal.GAS_G1000, geraetetyp=Geraetetyp.MULTIPLEXANLAGE
)


class TestGeraeteeigenschaften:
    @pytest.mark.parametrize(
        "geraeteeigenschaften, expected_json_dict",
        [
            pytest.param(
                example_geraeteeigenschaften,
                {"geraetemerkmal": "GAS_G1000", "geraetetyp": Geraetetyp.MULTIPLEXANLAGE},
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, geraeteeigenschaften: Geraeteeigenschaften, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Geraeteeigenschaften
        """
        assert_serialization_roundtrip(geraeteeigenschaften, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Geraeteeigenschaften()  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)

    @pytest.mark.parametrize(
        "not_a_geraetetyp",
        [
            pytest.param(17),  # not a geraetetyp
            pytest.param("foo"),  # not a geraetetyp
        ],
    )
    def test_failing_validation(self, not_a_geraetetyp: Any) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Geraeteeigenschaften(geraetemerkmal=Geraetemerkmal.GAS_G1000, geraetetyp=not_a_geraetetyp)

        assert "1 validation error" in str(excinfo.value)
        assert "value is not a valid enumeration member" in str(excinfo.value)



"""
Contains Hardware class
and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM
from bo4e.enum.geraetetyp import Geraetetyp

# pylint: disable=too-few-public-methods


class Hardware(COM):
    """
    Abbildung einer abrechenbaren Hardware

    .. raw:: html

        <object data="../_static/images/bo4e/com/Hardware.svg" type="image/svg+xml"></object>

    .. HINT::
        `Hardware JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Hardware.json>`_

    """

    # required attributes
    #: Eindeutiger Typ der Hardware
    geraetetyp: Geraetetyp
    #: Bezeichnung der Hardware
    bezeichnung: str



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Zeiteinheit(StrEnum):
    """
    Auflistung möglicher Einheiten zur Verwendung in zeitbezogenen Angaben.
    """

    SEKUNDE = "SEKUNDE"  #: Sekunde
    MINUTE = "MINUTE"  #: Minute
    STUNDE = "STUNDE"  #: Stunde
    VIERTEL_STUNDE = "VIERTEL_STUNDE"  #: Viertelstunde
    TAG = "TAG"  #: Tag
    WOCHE = "WOCHE"  #: Woche
    MONAT = "MONAT"  #: Monat
    QUARTAL = "QUARTAL"  #: Quartal
    HALBJAHR = "HALBJAHR"  #: Halbjahr
    JAHR = "JAHR"  #: Jahr



"""
Contains AufAbschlag class
and corresponding marshmallow schema for de-/serialization
"""

from typing import List, Optional

from pydantic import validator

from bo4e.com.com import COM
from bo4e.com.preisstaffel import Preisstaffel
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.aufabschlagstyp import AufAbschlagstyp
from bo4e.enum.aufabschlagsziel import AufAbschlagsziel
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from bo4e.validators import einheit_only_for_abschlagstyp_absolut

# pylint: disable=too-few-public-methods, too-many-instance-attributes


class AufAbschlag(COM):
    """
    Modell für die preiserhöhenden (Aufschlag) bzw. preisvermindernden (Abschlag) Zusatzvereinbarungen,
    die individuell zu einem neuen oder bestehenden Liefervertrag abgeschlossen wurden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/AufAbschlag.svg" type="image/svg+xml"></object>

    .. HINT::
        `AufAbschlag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/AufAbschlag.json>`_

    """

    # required attributes
    #: Bezeichnung des Auf-/Abschlags
    bezeichnung: str
    #: Werte für die gestaffelten Auf/Abschläge.
    staffeln: List[Preisstaffel]

    # optional attributes
    #: Beschreibung zum Auf-/Abschlag
    beschreibung: Optional[str] = None
    #: Typ des Aufabschlages (z.B. absolut oder prozentual).
    auf_abschlagstyp: Optional[AufAbschlagstyp] = None
    #: Diesem Preis oder den Kosten ist der Auf/Abschlag zugeordnet. Z.B. Arbeitspreis, Gesamtpreis etc..
    auf_abschlagsziel: Optional[AufAbschlagsziel] = None
    einheit: Optional[Waehrungseinheit] = None
    _einheit_check = validator("einheit", allow_reuse=True)(einheit_only_for_abschlagstyp_absolut)
    """ Gibt an in welcher Währungseinheit der Auf/Abschlag berechnet wird. Euro oder Ct..
    (Nur im Falle absoluter Aufschlagstypen). """
    #: Internetseite, auf der die Informationen zum Auf-/Abschlag veröffentlicht sind.
    website: Optional[str] = None
    #: Internetseite, auf der die Informationen zum Auf-/Abschlag veröffentlicht sind.
    gueltigkeitszeitraum: Optional[Zeitraum] = None
